import customtkinter as ctk

# This script contains functions to create widgets that are shared across GUIs.


def header_label(
    window,
    header_string,
    widget_cfg,
):
    """Header labels"""
    return ctk.CTkLabel(
        window,
        text=header_string,
        fg_color=widget_cfg["FG_COLOR"],
        text_color=widget_cfg["HEADER_TXT_COLOR"],
        font=(widget_cfg["HEADER_FONT_NAME"], widget_cfg["MAIN_HEADER_FONT_SIZE"]),
    )


def header_button(window, button_string, widget_cfg):
    """Header button"""
    return ctk.CTkButton(
        window,
        text=button_string,
        fg_color=widget_cfg["FG_COLOR"],
        hover_color=widget_cfg["HOVER_COLOR"],
        font=(widget_cfg["HEADER_FONT_NAME"], widget_cfg["HEADER_FONT_SIZE"]),
    )


def exit_button(window, widget_cfg):
    """Exit button"""
    return ctk.CTkButton(
        window,
        text="Exit",
        fg_color=widget_cfg["CLOSE_COLOR"],
        hover_color=widget_cfg["CLOSE_HOVER_COLOR"],
        font=(widget_cfg["HEADER_FONT_NAME"], widget_cfg["HEADER_FONT_SIZE"]),
    )


def label_and_entry_pair(
    window, label_text, entry_variable, widget_cfg, adv_cfg_textsize=False
):
    """Flexible label & entry widget-pair"""
    label = ctk.CTkLabel(
        window,
        text=label_text,
        font=(widget_cfg["TEXT_FONT_NAME"], widget_cfg["TEXT_FONT_SIZE"]),
    )
    entry = ctk.CTkEntry(
        window,
        textvariable=entry_variable,
        font=(widget_cfg["TEXT_FONT_NAME"], widget_cfg["TEXT_FONT_SIZE"]),
    )
    if adv_cfg_textsize:
        label.configure(
            font=(widget_cfg["TEXT_FONT_NAME"], widget_cfg["ADV_CFG_TEXT_FONT_SIZE"])
        )
        entry.configure(
            font=(widget_cfg["TEXT_FONT_NAME"], widget_cfg["ADV_CFG_TEXT_FONT_SIZE"])
        )
    return label, entry


def checkbox(window, label_text, entry_variable, widget_cfg, adv_cfg_textsize=False):
    """Checkbox"""
    checkbox = ctk.CTkCheckBox(
        window,
        text=label_text,
        variable=entry_variable,
        onvalue=True,
        offvalue=False,
        hover_color=widget_cfg["HOVER_COLOR"],
        fg_color=widget_cfg["FG_COLOR"],
        font=(widget_cfg["TEXT_FONT_NAME"], widget_cfg["TEXT_FONT_SIZE"]),
    )
    if adv_cfg_textsize:
        checkbox.configure(
            font=(widget_cfg["TEXT_FONT_NAME"], widget_cfg["ADV_CFG_TEXT_FONT_SIZE"])
        )
    return checkbox



from autogaita.resources.utils import try_to_run_gaita
import pandas as pd
import pandas.testing as pdt
import os
import shutil
import pdb
import pytest


# ............................  SIMI APPROVAL TESTS STRUCTURE  .........................
# 1. Run autogaita.simi for Subject (with the cfg used there)
# 2. Load the "Average Stepcycles".xlsx file from the repo and compare for
#    equivalence to average_data
# 3. Do the same for "Standard Devs. Stepcycle.xlsx" and std_data
# 4. Pass the test if the two df-pairs are equal


# ...............................  PREPARE - FOUR FIXTURES   ...........................


@pytest.fixture
def extract_true_dir():
    return "tests/test_data/universal3D_data/true_data/"


@pytest.fixture
def extract_info(tmp_path):
    info = {}
    info["name"] = "TestSubject"
    info["results_dir"] = os.path.join(tmp_path, info["name"])
    return info


@pytest.fixture
def extract_folderinfo():
    folderinfo = {}
    folderinfo["root_dir"] = "tests/test_data/universal3D_data/test_data/"
    folderinfo["sctable_filename"] = "SC Latency Table"
    folderinfo["postname_string"] = ""
    return folderinfo


@pytest.fixture
def extract_cfg():
    cfg = {}
    cfg["sampling_rate"] = 100
    cfg["dont_show_plots"] = True
    cfg["y_acceleration"] = True
    cfg["angular_acceleration"] = True
    cfg["bin_num"] = 25
    cfg["plot_SE"] = True
    cfg["normalise_height_at_SC_level"] = True
    cfg["plot_joint_number"] = 7
    cfg["legend_outside"] = True
    cfg["color_palette"] = "viridis"
    cfg["analyse_average_y"] = False
    cfg["joints"] = ["Midfoot", "Ankle", "Knee", "Hip", "Pelvis "]
    cfg["angles"] = {
        "name": ["Ankle", "Knee", "Hip"],
        "lower_joint": ["Midfoot", "Ankle", "Knee"],
        "upper_joint": ["Knee", "Hip", "Pelvis "],
    }
    return cfg


# ...............................  RUN - ONE APPROVAL TEST  ............................


@pytest.mark.slow  # https://docs.pytest.org/en/7.1.x/example/markers.html
def test_universal3D_approval(
    extract_true_dir, extract_info, extract_folderinfo, extract_cfg
):
    # run
    try_to_run_gaita(
        "Universal 3D", extract_info, extract_folderinfo, extract_cfg, False
    )
    # load true dfs from xlsx files
    true_av_df = pd.read_excel(
        os.path.join(
            extract_true_dir, extract_info["name"] + " - Average Stepcycle.xlsx"
        )
    )
    true_std_df = pd.read_excel(
        os.path.join(
            extract_true_dir, extract_info["name"] + " - Standard Devs. Stepcycle.xlsx"
        )
    )
    # test dfs are results of current run above
    test_av_df = pd.read_excel(
        os.path.join(
            extract_info["results_dir"],
            extract_info["name"] + " - Average Stepcycle.xlsx",
        )
    )
    test_std_df = pd.read_excel(
        os.path.join(
            extract_info["results_dir"],
            extract_info["name"] + " - Standard Devs. Stepcycle.xlsx",
        )
    )
    # finally assert equivalence of df-pairs
    pdt.assert_frame_equal(test_av_df, true_av_df)
    pdt.assert_frame_equal(test_std_df, true_std_df)



from autogaita.group.group_3_PCA import run_PCA, convert_PCA_bins_to_list
from autogaita.group.group_4_stats import run_ANOVA
import pytest
from sklearn import datasets
import pandas as pd
import numpy as np
import math


# %%................................  fixtures  ........................................
@pytest.fixture
def extract_folderinfo(tmp_path):
    return {
        "group_names": ["group1", "group2"],
        "group_dirs": ["/path/to/group1", "/path/to/group2"],
        "results_dir": tmp_path,
    }


@pytest.fixture
def extract_cfg():
    return {
        "do_permtest": True,
        "do_anova": True,
        "permutation_number": 100,
        "PCA_n_components": 3,
        "PCA_custom_scatter_PCs": "",
        "PCA_save_3D_video": False,  # True
        "PCA_bins": "",
        "stats_threshold": 0.05,
        "plot_SE": False,
        "color_palette": "viridis",
        "dont_show_plots": True,
        "legend_outside": True,
        "which_leg": "left",
        "anova_design": "Mixed ANOVA",
        "PCA_variables": [],
        "stats_variables": [],
    }


# %%..............................  statistics  ........................................


# NOTE !!!
# => This test is outdated and wrong because it is for two within subject factors (for
#    example 3 separate behavioural tests and pre/post medication treatment or so -
#    done by each subject)
# => I will not delete this since I might want to re-use it if I should support such
#    designs
#
# def test_twoway_RM_ANOVA(extract_cfg):
#     # Adopted example data from https://real-statistics.com/
#     # See the RM ANOVA xlsx file for complete link
#     extract_cfg["do_anova"] = True
#     extract_cfg["anova_design"] = "RM ANOVA"
#     stats_df = pd.read_excel("tests/test_data/group_data/RM ANOVA Example Data.xlsx")
#     stats_var = "Value"
#     result = run_ANOVA(stats_df, stats_var, extract_cfg)
#     # Note that the last 2 assert statements have a different tolerance because those
#     # p-values differed a bit. Not sure why but it's a tiny amount and tolerable IMO
#     pytest.set_trace()
#     assert math.isclose(result["p-unc"][0], stats_df["p(A)"][0], abs_tol=1e-05)
#     assert math.isclose(result["p-unc"][1], stats_df["p(B)"][0], abs_tol=1e-05)
#     assert math.isclose(result["p-unc"][2], stats_df["p(AxB)"][0], abs_tol=1e-05)
#     assert math.isclose(result["p-GG-corr"][0], stats_df["GG-p(A)"][0], abs_tol=1e-05)
#     assert math.isclose(result["p-GG-corr"][1], stats_df["GG-p(B)"][0], abs_tol=1e-04)
#     assert math.isclose(result["p-GG-corr"][2], stats_df["GG-p(AxB)"][0], abs_tol=1e-02)


def test_Mixed_ANOVA(extract_cfg):
    # Adopted example data from https://real-statistics.com/
    # See the Mixed ANOVA xlsx file for complete link
    extract_cfg["do_anova"] = True
    extract_cfg["anova_design"] = "Mixed ANOVA"
    stats_df = pd.read_excel("tests/test_data/group_data/Mixed ANOVA Example Data.xlsx")
    stats_var = "Value"
    result = run_ANOVA(stats_df, stats_var, extract_cfg)
    # no GG p vals here
    assert math.isclose(result["p-unc"][0], stats_df["p(A)"][0], abs_tol=1e-05)
    assert math.isclose(result["p-unc"][1], stats_df["p(B)"][0], abs_tol=1e-05)
    assert math.isclose(result["p-unc"][2], stats_df["p(AxB)"][0], abs_tol=1e-05)


# %%..................................  PCA  ...........................................
def test_run_PCA(extract_cfg):
    # Replicate the example found in https://www.kdnuggets.com/2023/05/
    # principal-component-analysis-pca-scikitlearn.html using our PCA df and PCA_info
    # structure
    # fmt: off
    true_results_PCA_eigenvectors = [[0.1443294, -0.24518758, -0.00205106, -0.23932041,  0.14199204,  0.39466085, 0.4229343, -0.2985331, 0.31342949, -0.0886167,0.29671456,  0.37616741, 0.28675223],
    [-0.48365155, -0.22493093, -0.31606881,  0.0105905,  -0.299634,   -0.06503951
    , 0.00335981, -0.02877949, -0.03930172, -0.52999567,  0.27923515,  0.16449619,
    -0.36490283],
    [-0.20738262,  0.08901289,  0.6262239,   0.61208035,  0.13075693,  0.14617896,
    0.1506819,   0.17036816,  0.14945431, -0.13730621,  0.08522192,  0.16600459,
    -0.12674592]]
    true_results_PCA_explained_var = [0.36198848, 0.1920749 , 0.11123631]
    # fmt: on
    wine_data = datasets.load_wine(as_frame=True)
    wine_df = wine_data.data
    features = wine_df.columns
    PCA_df, PCA_info = run_PCA(wine_df, features, extract_cfg)
    for i in range(3):
        # absolute values are compared because the signs can be different w. eigenvecs
        assert np.allclose(
            np.absolute(PCA_info["eigenvectors"][i]),
            np.absolute(true_results_PCA_eigenvectors[i]),
            atol=1e-05,
        )
    assert np.allclose(
        PCA_info["explained_vars"], true_results_PCA_explained_var, atol=1e-05
    )


cases = (
    (
        "1-30,50,70-100",
        100,
        np.arange(1, 31).tolist() + [50] + np.arange(70, 101).tolist(),
    ),
    ("1-70", 25, [4, 8, 12, 16, 20, 24, 28, 32, 36, 40, 44, 48, 52, 56, 60, 64, 68]),
    ("2, 10, 22, 44, 50, 79", 50, [2, 10, 22, 44, 50]),
    ("1-50,70-80", 10, [10, 20, 30, 40, 50, 70, 80]),
)  # fmt: skip
@pytest.mark.parametrize("PCA_bins, bin_num, expected_bins_list", cases)
def test_convert_PCA_bins_to_list(
    PCA_bins, bin_num, expected_bins_list, extract_folderinfo, extract_cfg
):
    extract_cfg["PCA_bins"] = PCA_bins
    extract_cfg["bin_num"] = bin_num
    updated_cfg = convert_PCA_bins_to_list(extract_folderinfo, extract_cfg)
    assert np.array_equal(updated_cfg["PCA_bins"], np.array(expected_bins_list))



from autogaita.resources.utils import try_to_run_gaita
import os
import tkinter as tk


def prepare_DLC_df(df, separator=" "):
    """Prepare the DLC dataframe after loading w.r.t. column names & df-index
    Note
    ----
    separator is used in universal3D_datafile_preparation
    """
    new_column_strings = list()  # data df
    for j in range(df.shape[1]):
        new_column_strings.append(df.iloc[0, j] + separator + df.iloc[1, j])
    df.columns = new_column_strings
    # next lines indices are because: scorer row becomes the column, bodypart row is row
    # 0, coords row is row 1 and we thus include row 2 onwards. col 1 onwards is obvious
    df = df.iloc[2:, 1:]
    df.index = range(len(df))
    df = df.astype(float)
    return df


def find_number(fullstring, prestring, poststring):
    """Find (mouse/run) number based on user-defined strings in filenames"""
    start_idx = fullstring.find(prestring) + len(prestring)
    end_idx = fullstring.find(poststring)
    return int(fullstring[start_idx:end_idx])



# %% imports
from autogaita.resources.utils import write_issues_to_textfile
import os
import json
import matplotlib.pyplot as plt
import seaborn as sns

# %% constants
from autogaita.resources.constants import ISSUES_TXT_FILENAME, CONFIG_JSON_FILENAME
from autogaita.group.group_constants import (
    GROUP_CONFIG_TXT_FILENAME,
    STATS_TXT_FILENAME,
    MULTCOMP_EXCEL_FILENAME_1,
    MULTCOMP_EXCEL_FILENAME_2,
    ORIG_SHEET_NAME,
    CONTRAST_SPLIT_STR,
)


# %% .................  workflow step #1 - unpack & prepare vars  ....................


def some_prep(folderinfo, cfg):
    """Add some folderinfo & cfg variables to the dictionaries for further processes"""

    # unpack
    group_names = folderinfo["group_names"]
    group_dirs = folderinfo["group_dirs"]
    results_dir = folderinfo["results_dir"]

    # names & paths to folderinfo
    if not os.path.isdir(results_dir):
        os.makedirs(results_dir)
    folderinfo["contrasts"] = []
    for i in range(len(group_names)):
        for j in range(i + 1, len(group_names)):
            contrast = group_names[i] + CONTRAST_SPLIT_STR + group_names[j]
            folderinfo["contrasts"].append(contrast)

    # check if we previously had saved info files, if so delete them
    for info_file_name in [
        GROUP_CONFIG_TXT_FILENAME,
        ISSUES_TXT_FILENAME,
        STATS_TXT_FILENAME,
        MULTCOMP_EXCEL_FILENAME_1,
        MULTCOMP_EXCEL_FILENAME_2,
    ]:
        info_file_path = os.path.join(results_dir, info_file_name)
        if os.path.exists(info_file_path):
            os.remove(info_file_path)

    # extracted_cfg_vars: save_to_xls, PCA stuff & dont show plots
    cfg = extract_cfg_vars(folderinfo, cfg)

    # see if there's a config json file and add to cfg dict
    for g_idx, group_dir in enumerate(group_dirs):
        with open(
            os.path.join(group_dir, CONFIG_JSON_FILENAME), "r"
        ) as config_json_file:
            config_vars_from_json = json.load(config_json_file)
            for key in config_vars_from_json.keys():
                # assigning like this ensure all keys are in all jsons across groups
                if g_idx == 0:
                    cfg[key] = config_vars_from_json[key]
                else:
                    # sanity check for group-differences in cfg variables!
                    if (key not in cfg.keys()) | (
                        cfg[key] != config_vars_from_json[key]
                    ):
                        error_message = (
                            "config.json variables differ between groups!"
                            + "\nPlease make sure that all cfg variables between "
                            + "groups match & try again!"
                        )
                        raise ValueError(error_message)
                    else:
                        cfg[key] = config_vars_from_json[key]

    # after having checked cfg keys for equivalence, we have to make sure that
    # hind_joints is renamed to joints if DLC
    # => note that cfg will not be updated (and overwritten from input if Universal 3D so it's)
    #    okay that we do this
    if cfg["tracking_software"] == "DLC":
        cfg["joints"] = cfg["hind_joints"]
    joints = cfg["joints"]
    angles = cfg["angles"]

    # prepare some plotting color stuff
    cfg["group_color_cycler"] = plt.cycler(
        "color", sns.color_palette(cfg["color_palette"], len(group_names))
    )
    cfg["group_color_dict"] = dict(
        zip(group_names, cfg["group_color_cycler"].by_key()["color"])
    )
    cfg["joint_color_cycler"] = plt.cycler(
        "color", sns.color_palette(cfg["color_palette"], len(joints))
    )
    cfg["angle_color_cycler"] = plt.cycler(
        "color", sns.color_palette(cfg["color_palette"], len(angles["name"]))
    )

    return folderinfo, cfg


def extract_cfg_vars(folderinfo, cfg):
    """Extract save_to_xls from example Normalised dfs and sanity check
    that they match between groups. Also some stuff for PCA!
    """

    group_names = folderinfo["group_names"]
    group_dirs = folderinfo["group_dirs"]

    # ................................  save_to_xls  ...................................
    save_to_xls = [None] * len(group_dirs)
    for g, group_dir in enumerate(group_dirs):
        all_results_folders = os.listdir(
            group_dir
        )  # remove no-results valid_results_folders
        valid_results_folders = []
        # => Note if there's ambiguity / mixed filetypes, we set save_to_xls to True!
        sheet_type_mismatch_message = (
            "\n***********\n! WARNING !\n***********\n"
            + "Mismatch in sheet file types for group "
            + group_names[g]
            + "!\nSaving all output sheets to"
            + ".xlsx!\nRe-run first level & only save .csvs if "
            + "you want .csv files of group results!"
        )
        for folder in all_results_folders:
            # create save_to_xls here, there are two cases we have to deal with:
            # case 1: we found a csv file
            if os.path.exists(
                os.path.join(
                    group_dir,
                    folder,
                    folder + " - " + ORIG_SHEET_NAME + ".csv",
                )
            ):
                valid_results_folders.append(folder)
                if save_to_xls[g] is None:
                    save_to_xls[g] = False
                if save_to_xls[g] is True:
                    print(sheet_type_mismatch_message)
                    write_issues_to_textfile(sheet_type_mismatch_message, folderinfo)
            # case 2: we found a xlsx file
            elif os.path.exists(
                os.path.join(
                    group_dir,
                    folder,
                    folder + " - " + ORIG_SHEET_NAME + ".xlsx",
                )
            ):
                valid_results_folders.append(folder)
                if save_to_xls[g] is None:
                    save_to_xls[g] = True
                if save_to_xls[g] is False:
                    save_to_xls[g] = True
                    print(sheet_type_mismatch_message)
                    write_issues_to_textfile(sheet_type_mismatch_message, folderinfo)
        # test that at least 1 folder has valid results for all groups
        if not valid_results_folders:
            no_valid_results_error = (
                "\n*********\n! ERROR !\n*********\n"
                + "No valid results folder found for "
                + group_names[g]
                + "\nFix & re-run!"
            )
            print(no_valid_results_error)
            write_issues_to_textfile(no_valid_results_error, folderinfo)
    # assign to our cfg dict after group loop
    cfg["save_to_xls"] = save_to_xls

    # .........................  test if PCA config is valid  ..........................
    # only test if user wants PCA (ie. selected any features) and is not using the
    # var-explained appproiach
    if cfg["PCA_variables"] and cfg["PCA_n_components"] > 1:
        if len(cfg["PCA_variables"]) < cfg["PCA_n_components"]:
            PCA_variable_num = len(cfg["PCA_variables"])
            PCA_variables_str = "\n".join(cfg["PCA_variables"])
            PCA_error_message = (
                "\n*********\n! ERROR !\n*********\n"
                + "\nPCA Configuration invalid, number of input features must exceed "
                + "the number of principal components to compute!\n"
                + str(PCA_variable_num)
                + " PCA variables: \n"
                + PCA_variables_str
                + "\n & Number of wanted PCs: "
                + str(cfg["PCA_n_components"])
                + "\n Fix & re-run!"
            )
            write_issues_to_textfile(PCA_error_message, folderinfo)
            raise ValueError(PCA_error_message)
        if cfg["PCA_n_components"] < 2:
            print(
                "\n***********\n! WARNING !\n***********\n"
                + "Number of principal components of PCA cannot be 0 or 1!"
                + "\nRunning PCA on 2 components - if you do not want to perform PCA, "
                + "just don't choose any variables for it."
            )
            cfg["PCA_n_components"] = 2  # make sure to update in cfg dict
    # small fix to ensure that PCA vars don't have duplicates (this would have severe
    # consequences for run_PCA since features of PCA_model output and "my" features var
    # would not match)
    # => it should never happen via GUI really but I did this myself using group_dlcrun #    while not paying attention...
    unique_PCA_vars = list(set(cfg["PCA_variables"]))
    if len(unique_PCA_vars) != len(cfg["PCA_variables"]):
        duplicate_vars = [
            var for var in cfg["PCA_variables"] if cfg["PCA_variables"].count(var) > 1
        ]
        PCA_duplicates_error_message = (
            "\n*********\n! WARNING !\n*********\n"
            + "\nWe found duplicates in your PCA variables list!"
            + "\nWe removed them for you!"
            + "\n\nDuplicate variables were:\n"
            + "\n".join(list(set(duplicate_vars)))
        )
        print(PCA_duplicates_error_message)
        write_issues_to_textfile(PCA_duplicates_error_message, folderinfo)
        cfg["PCA_variables"] = unique_PCA_vars
    # check if PCA bin num is valid, two tests and 1 fix
    # => note there is separate code that transforms this string input into an int-list
    if cfg["PCA_bins"]:
        PCA_bins_error_message = ""
        cfg["PCA_bins"] = cfg["PCA_bins"].replace(" ", "")  # remove spaces
        # test 1: only digits, commas and dashes allowed
        for char in cfg["PCA_bins"]:
            if char.isdigit() or char in ["-", ","]:
                pass
            else:
                PCA_bins_error_message = (
                    "\n*********\n! ERROR !\n*********\n"
                    + "\nPCA Bin Number invalid!"
                    + "\nPlease use only numbers, commas and dashes in your input!"
                    + "\nFix & re-run!"
                )
        # test 2: no comma next to dash
        if ",-" in cfg["PCA_bins"] or "-," in cfg["PCA_bins"]:
            PCA_bins_error_message = (
                "\n*********\n! ERROR !\n*********\n"
                + "\nPCA Bin Number invalid!"
                + "\nCommas cannot be next to dashes in your input!"
                + "\nFix & re-run!"
            )
        # if any test failed we raise an error
        # => note we could continue execution without custom PCA bins but if the user
        #    tried to set them, it's likely important to them to have custom bins
        #    so we just stop everything and help them fix it
        if PCA_bins_error_message:
            print(PCA_bins_error_message)
            write_issues_to_textfile(PCA_bins_error_message, folderinfo)
            raise ValueError(PCA_bins_error_message)
        # fix for users: remove first and last characters if not digits
        while not cfg["PCA_bins"][0].isdigit():
            cfg["PCA_bins"] = cfg["PCA_bins"][1:]
        while not cfg["PCA_bins"][-1].isdigit():
            cfg["PCA_bins"] = cfg["PCA_bins"][:-1]

    return cfg



from autogaita.resources.utils import try_to_run_gaita
import os


# %% main function


def universal3D_multirun():
    """
    Batchrun script to run AutoGaitA Universal 3D for a folder of datasets.
    folderinfo & cfg dictionaries must be configured as explained in our documentation. See the "AutoGaitA without the GUI" section of our documentation for references to in-depth explanations to all dictionary keys (note that each key of dicts corresponds to some object in the AutoGaitA Universal 3D GUI)
    """
    # folderinfo
    folderinfo = {}
    folderinfo["root_dir"] = "/Users/mahan/sciebo/Research/AutoGaitA/Human/Testing2/"
    folderinfo["results_dir"] = ""
    folderinfo["sctable_filename"] = "SC Latency Table"
    folderinfo["postname_string"] = ""
    # cfg
    cfg = {}
    cfg["sampling_rate"] = 100  # base cfg
    cfg["dont_show_plots"] = True
    cfg["y_acceleration"] = True
    cfg["angular_acceleration"] = True
    cfg["bin_num"] = 25
    cfg["plot_SE"] = True
    cfg["normalise_height_at_SC_level"] = True
    cfg["plot_joint_number"] = 7
    cfg["color_palette"] = "viridis"
    cfg["legend_outside"] = True
    cfg["analyse_average_y"] = True
    cfg["joints"] = ["Midfoot", "Ankle", "Knee", "Hip", "Pelvis", "Shoulder", "Neck"]
    cfg["angles"] = {
        "name": ["Ankle", "Knee"],
        "lower_joint": ["Midfoot", "Ankle"],
        "upper_joint": ["Knee", "Hip"],
    }
    # run a single gaita run for each entry of info
    info = extract_info(folderinfo)
    for idx, name in enumerate(info["name"]):
        run_singlerun(idx, info, folderinfo, cfg)


# %% local functions


def run_singlerun(idx, info, folderinfo, cfg):
    """Run the main code of individual run-analyses based on current cfg"""
    # extract and pass info of this mouse/run (also update resdir)
    this_info = {}
    keynames = info.keys()
    for keyname in keynames:
        this_info[keyname] = info[keyname][idx]
    # important to only pass this_info to main script here (1 run at a time!)
    try_to_run_gaita("Universal 3D", this_info, folderinfo, cfg, True)


def extract_info(folderinfo):
    """Prepare a dict of lists that include unique name infos"""
    root_dir = folderinfo["root_dir"]
    results_dir = folderinfo["results_dir"]
    sctable_filename = folderinfo["sctable_filename"]
    postname_string = folderinfo["postname_string"]
    info = {"name": [], "results_dir": []}
    for filename in os.listdir(root_dir):
        # dont try to combine the two "join" if blocks into one - we want to append
        # results dir WHENEVER we append name!
        if not postname_string:
            # dont use endswith below to catch .xlsx too
            if (".xls" in filename) & (sctable_filename not in filename):
                info["name"].append(filename.split(".xls")[0])
                if results_dir:
                    info["results_dir"].append(
                        os.path.join(results_dir, info["name"][-1])
                    )
                else:
                    info["results_dir"].append(
                        os.path.join(root_dir, "Results", info["name"][-1])
                    )
        else:
            if postname_string in filename:
                info["name"].append(filename.split(postname_string)[0])
                if results_dir:
                    info["results_dir"].append(
                        os.path.join(results_dir, info["name"][-1])
                    )
                else:
                    info["results_dir"].append(
                        os.path.join(root_dir, "Results", info["name"][-1])
                    )
    return info


# %% what happens if we just hit run
if __name__ == "__main__":
    universal3D_multirun()



import os
import json
import tkinter as tk
from threading import Thread
from autogaita.common2D.common2D_utils import run_singlerun_in_multirun, extract_info
from autogaita.resources.utils import try_to_run_gaita

# %%..........  functions preparing and calling runs of 2D autogaita  ..................


def runanalysis(tracking_software, analysis, this_runs_results, this_runs_cfg):
    """Run the main program"""
    if analysis == "single":
        run_thread = Thread(
            target=analyse_single_run,
            args=(tracking_software, this_runs_results, this_runs_cfg),
        )
    elif analysis == "multi":
        run_thread = Thread(
            target=analyse_multi_run,
            args=(tracking_software, this_runs_results, this_runs_cfg),
        )
    run_thread.start()


def analyse_single_run(tracking_software, this_runs_results, this_runs_cfg):
    """Prepare for one execution of autogaita_dlc & execute"""
    # prepare (folderinfo first to handle root_dir "/" issue if needed)
    folderinfo = prepare_folderinfo(tracking_software, this_runs_results)
    if folderinfo is None:
        error_msg = (
            "No directory found at: " + this_runs_results["root_dir"] + " - try again!"
        )
        tk.messagebox.showerror(title="Try again", message=error_msg)
        print(error_msg)
        return
    info = {}  # info dict: run-specific info
    if tracking_software == "DLC":
        info["mouse_num"] = this_runs_results["mouse_num"]
        info["run_num"] = this_runs_results["run_num"]
        info["name"] = "ID " + str(info["mouse_num"]) + " - Run " + str(info["run_num"])
    elif tracking_software == "SLEAP":
        info["name"] = this_runs_results["name"]
    if this_runs_cfg["results_dir"]:
        info["results_dir"] = os.path.join(this_runs_cfg["results_dir"], info["name"])
    else:
        info["results_dir"] = os.path.join(
            folderinfo["root_dir"], "Results", info["name"]
        )
    # execute
    try_to_run_gaita(tracking_software, info, folderinfo, this_runs_cfg, False)


def analyse_multi_run(tracking_software, this_runs_results, this_runs_cfg):
    """Prepare for multi-execution of autogaita_dlc & loop-execute"""
    # prepare (folderinfo first to handle root_dir "/" issue if needed)
    folderinfo = prepare_folderinfo(tracking_software, this_runs_results)
    if folderinfo is None:
        error_msg = (
            "No directory found at: " + this_runs_results["root_dir"] + " - try again!"
        )
        tk.messagebox.showerror(title="Try again", message=error_msg)
        print(error_msg)
        return
    # folderinfo has info of individual runs - extract individual runs info
    # => different workflows here for DLC and SLEAP
    info = extract_info(tracking_software, folderinfo, in_GUI=True)
    # this can fail for DLC which will return None - use this to handle the error and
    # stop execution
    if info is None:
        return
    # if there was no error, loop through individual runs
    for idx in range(len(info["name"])):
        run_singlerun_in_multirun(
            tracking_software, idx, info, folderinfo, this_runs_cfg
        )


def prepare_folderinfo(tracking_software, this_runs_results):
    # IMPORTANT NOTE
    # This function should only return None if there are issues with root_dir
    # since we handle the error in outside functions because of this (i.e. the meaning
    # behind the None should NOT change in the future!)
    if len(this_runs_results["root_dir"]) == 0:
        return
    folderinfo = {}
    folderinfo["root_dir"] = this_runs_results["root_dir"]
    # to make sure root_dir works under windows
    # (windows is okay with all dir-separators being "/", so make sure it is!)
    folderinfo["root_dir"] = folderinfo["root_dir"].replace(os.sep, "/")
    if folderinfo["root_dir"][-1] != "/":
        folderinfo["root_dir"] = folderinfo["root_dir"] + "/"
    if not os.path.exists(folderinfo["root_dir"]):
        return
    folderinfo["sctable_filename"] = this_runs_results["sctable_filename"]
    folderinfo["data_string"] = this_runs_results["data_string"]
    folderinfo["beam_string"] = this_runs_results["beam_string"]
    if tracking_software == "DLC":
        folderinfo["premouse_string"] = this_runs_results["premouse_string"]
        folderinfo["postmouse_string"] = this_runs_results["postmouse_string"]
        folderinfo["prerun_string"] = this_runs_results["prerun_string"]
        folderinfo["postrun_string"] = this_runs_results["postrun_string"]
    return folderinfo


# %%...............  important function getting values from tk vars  ...................


def get_results_and_cfg(results, cfg, analysis, variable_dict):
    """Before calling analysis, use .get() to extract values from tk-vars"""

    # unpack var dict
    FLOAT_VARS = variable_dict["FLOAT_VARS"]
    INT_VARS = variable_dict["INT_VARS"]
    LIST_VARS = variable_dict["LIST_VARS"]
    DICT_VARS = variable_dict["DICT_VARS"]

    # 1) We make sure to return and use for all "run"-purposes (i.e. everything
    #    that follows this local function) "this_" dicts (stuff is dangerous
    #    otherwise!)
    # 2) I implemented the usage & transformation of FLOAT_&INT_VARS the way I
    #    do here (instead of initialising as (e.g.) tk.IntVar because otherwise
    #    tkinter complains whenever an Entry is empty @ any time, just because
    #    the user is inputting their values.
    output_dicts = [{}, {}]
    for i in range(len(output_dicts)):
        if i == 0:
            input_dict = results
        elif i == 1:
            input_dict = cfg
        for key in input_dict.keys():
            # exclude mouse_num and run_num from int conversion
            # in case analysis is multi
            if analysis == "multi" and (key == "mouse_num" or key == "run_num"):
                output_dicts[i][key] = input_dict[key].get()
            elif key in FLOAT_VARS:
                # make sure that if user doesn't want to convert then we just use 0 so
                # the float() statement doesn't throw an error
                # => set the outer dict directly instead of modulating the input_dict,
                #    since that would also modify cfg which might be dangerous for
                #    future runs
                if (key == "pixel_to_mm_ratio") & (cfg["convert_to_mm"].get() is False):
                    output_dicts[i][key] = 0
                else:
                    try:
                        output_dicts[i][key] = float(input_dict[key].get())
                    except ValueError:
                        return (
                            key
                            + " value was not a decimal number, but "
                            + input_dict[key].get()
                        )
            elif key in INT_VARS:
                try:
                    output_dicts[i][key] = int(input_dict[key].get())
                except ValueError:
                    return (
                        key
                        + " value was not an integer number, but "
                        + input_dict[key].get()
                    )
            elif key in LIST_VARS:
                # if list of strings, initialise output empty list and get & append vals
                output_dicts[i][key] = []
                for list_idx in range(len(input_dict[key])):
                    output_dicts[i][key].append(input_dict[key][list_idx].get())
            elif key in DICT_VARS:
                # if dict of list of strings, initialise as empty dict and assign stuff
                # key = "angles" or other DICT_VAR
                # inner_key = "name" & "lower_ / upper_joint"
                # list_idx = idx of list of strings of inner_key
                output_dicts[i][key] = {}
                for inner_key in input_dict[key]:
                    output_dicts[i][key][inner_key] = []
                    for list_idx in range(len(input_dict[key][inner_key])):
                        output_dicts[i][key][inner_key].append(
                            input_dict[key][inner_key][list_idx].get()
                        )
            else:
                output_dicts[i][key] = input_dict[key].get()
    this_runs_results = output_dicts[0]
    this_runs_cfg = output_dicts[1]
    return this_runs_results, this_runs_cfg


# ............................ gui config file operations  ............................
def update_config_file(
    results,
    cfg,
    autogaita_folder_path,
    config_file_name,
    list_vars,
    dict_vars,
    tk_str_vars,
    tk_bool_vars,
):
    """updates the dlc_gui_config file with this runs parameters"""
    # transform tkVars into normal strings and bools
    output_dicts = [{}, {}]
    for i in range(len(output_dicts)):
        if i == 0:
            # in case update_config_file is called before results is defined
            # as in the creation of the exit_button in the dlc_gui() function
            # the results dict of the last run is used and only cfg is updated
            if results == "results dict not defined yet":
                # runwindow = None as we dont need the tk.Vars to refer to a specific window
                input_dict = extract_results_from_json_file(
                    None,
                    autogaita_folder_path,
                    config_file_name,
                    tk_str_vars,
                    tk_bool_vars,
                )
            else:
                input_dict = results
        elif i == 1:
            input_dict = cfg
        for key in input_dict.keys():
            if key in list_vars:
                # if list of strings, initialise output empty list and get & append vals
                output_dicts[i][key] = []
                for list_idx in range(len(input_dict[key])):
                    output_dicts[i][key].append(input_dict[key][list_idx].get())
            elif key in dict_vars:
                # if dict of list of strings, initialise as empty dict and assign stuff
                # key = "angles" or other DICT_VAR
                # inner_key = "name" & "lower_ / upper_joint"
                # list_idx = idx of list of strings of inner_key
                output_dicts[i][key] = {}
                for inner_key in input_dict[key]:
                    output_dicts[i][key][inner_key] = []
                    for list_idx in range(len(input_dict[key][inner_key])):
                        output_dicts[i][key][inner_key].append(
                            input_dict[key][inner_key][list_idx].get()
                        )
            else:
                output_dicts[i][key] = input_dict[key].get()

    # merge the two configuration dictionaries
    configs_list = [output_dicts[0], output_dicts[1]]  # 0 = results, 1 = cfg, see above
    # write the configuration file
    with open(
        os.path.join(autogaita_folder_path, config_file_name), "w"
    ) as config_json_file:
        json.dump(configs_list, config_json_file, indent=4)


def extract_cfg_from_json_file(
    window,
    autogaita_folder_path,
    config_file_name,
    list_vars,
    dict_vars,
    tk_str_vars,
    tk_bool_vars,
):
    """loads the cfg dictionary from the config file"""
    # load the configuration file
    with open(
        os.path.join(autogaita_folder_path, config_file_name), "r"
    ) as config_json_file:
        # config_json contains list with 0 -> result and 1 -> cfg data
        last_runs_cfg = json.load(config_json_file)[1]

    cfg = {}
    # assign values to the cfg dict
    for key in last_runs_cfg.keys():
        if key in tk_bool_vars:
            cfg[key] = tk.BooleanVar(window, last_runs_cfg[key])
        elif key in list_vars:
            cfg[key] = []
            for entry in last_runs_cfg[key]:
                cfg[key].append(tk.StringVar(window, entry))
        elif key in dict_vars:
            cfg[key] = {}
            for subkey in last_runs_cfg[key]:
                cfg[key][subkey] = []
                for entry in last_runs_cfg[key][subkey]:
                    cfg[key][subkey].append(tk.StringVar(window, entry))
        elif key in tk_str_vars:  # Integers are also saved as strings
            cfg[key] = tk.StringVar(window, last_runs_cfg[key])
    return cfg


def extract_results_from_json_file(
    window, autogaita_folder_path, config_file_name, tk_str_vars, tk_bool_vars
):
    """loads the results dictionary from the config file"""

    # load the configuration file
    with open(
        os.path.join(autogaita_folder_path, config_file_name), "r"
    ) as config_json_file:
        # config_json contains list with 0 -> result and 1 -> cfg data
        last_runs_results = json.load(config_json_file)[0]

    results = {}
    for key in last_runs_results.keys():
        if key in tk_str_vars:
            results[key] = tk.StringVar(window, last_runs_results[key])
        elif key in tk_bool_vars:
            results[key] = tk.BooleanVar(window, last_runs_results[key])

    return results



import autogaita


def group_dlcrun():
    """
    Batchrun script to run AutoGaitA Group for Results obtained with AutoGaitA DLC.
    folderinfo & cfg dictionaries must be configured as explained in our documentation. See the "AutoGaitA without the GUI" section of our documentation for references to in-depth explanations to all dictionary keys (note that each key of dicts corresponds to some object in the AutoGaitA Group GUI)
    """
    # folderinfo
    # => Note that length of folderinfo's group_names & group_dirs lists determines how #    many groups are compared.
    # => Also note that indices must correspond (i.e., idx #    1's name will be used #    for dataset stored in group_dir's idx 1)
    folderinfo = {}
    folderinfo["group_names"] = [
        # ----------
        # RM Dataset
        # "5mm",
        # "12mm",
        # "25mm",
        # -------
        # Interaction Issue Dataset
        # "Control",
        # "Silenced",
        # --------
        # Testing Dataset
        # "one",
        # "two",
        # --------
        # Beam x Age for Manuscript
        "3m_5mm",
        "3m_12mm",
        "3m_25mm",
        "24m_5mm",
        "24m_12mm",
        "24m_25mm",
    ]
    folderinfo["group_dirs"] = [
        # --------------------------
        # Full Dataset (39 subjects)
        # "/Users/mahan/sciebo/Research/AutoGaitA/Mouse/Sanity Test Interaction Graziana/Result 10 joints/Control Full",
        # "/Users/mahan/sciebo/Research/AutoGaitA/Mouse/Sanity Test Interaction Graziana/Result 10 joints/Silenced Full/",
        # --------------------------
        # Reduced Dataset (5 subjects)
        # "/Users/mahan/sciebo/Research/AutoGaitA/Mouse/Sanity Test Interaction Graziana/Result 10 joints/Control",
        # "/Users/mahan/sciebo/Research/AutoGaitA/Mouse/Sanity Test Interaction Graziana/Result 10 joints/Silenced/",
        # --------------------------
        # Testing Dataset
        # "/Users/mahan/sciebo/Research/AutoGaitA/Mouse/Testing/Group 1/",
        # "/Users/mahan/sciebo/Research/AutoGaitA/Mouse/Testing/Group 2/",
        # --------------------------
        # RM Dataset (example data in repo)
        # "/Users/mahan/sciebo/PythonCode/gaita_repo/example data/5mm/Results/",
        # "/Users/mahan/sciebo/PythonCode/gaita_repo/example data/12mm/Results/",
        # "/Users/mahan/sciebo/PythonCode/gaita_repo/example data/25mm/Results/",
        # --------------------------
        # Beam x Age for Manuscript
        "/Users/mahan/sciebo/Research/AutoGaitA/Mouse/Full Dataset and Results Nov 2024/5mm/9w",
        "/Users/mahan/sciebo/Research/AutoGaitA/Mouse/Full Dataset and Results Nov 2024/12mm/9w",
        "/Users/mahan/sciebo/Research/AutoGaitA/Mouse/Full Dataset and Results Nov 2024/25mm/9w",
        "/Users/mahan/sciebo/Research/AutoGaitA/Mouse/Full Dataset and Results Nov 2024/5mm/104w",
        "/Users/mahan/sciebo/Research/AutoGaitA/Mouse/Full Dataset and Results Nov 2024/12mm/104w",
        "/Users/mahan/sciebo/Research/AutoGaitA/Mouse/Full Dataset and Results Nov 2024/25mm/104w"
    ]
    folderinfo["results_dir"] = (
        # "/Users/mahan/sciebo/Research/AutoGaitA/Mouse/example_data_results/"
        # "/Users/mahan/sciebo/Research/AutoGaitA/Mouse/Sanity Test Interaction Graziana/Result 10 joints/Mahan Results/"
        # "/Users/mahan/sciebo/Research/AutoGaitA/Mouse/Testing/GroupResults/"
        "/Users/mahan/sciebo/Research/AutoGaitA/Mouse/Full Dataset and Results Nov 2024/YoungOld3BeamsForPaper"
    )
    # cfg
    cfg = {}
    cfg["do_permtest"] = False  # True
    cfg["do_anova"] = True
    cfg["permutation_number"] = 100
    cfg["PCA_n_components"] = 0.9
    cfg["PCA_custom_scatter_PCs"] = "1,3;2,3;2,3,4;;;;"
    cfg["PCA_save_3D_video"] = False
    cfg["PCA_bins"] = ""  # "0-10,24,50-75"
    cfg["stats_threshold"] = 0.05
    cfg["plot_SE"] = True
    cfg["color_palette"] = "inferno"
    cfg["dont_show_plots"] = True
    cfg["legend_outside"] = True
    cfg["which_leg"] = "left"
    cfg["anova_design"] = "RM ANOVA"
    # cfg["anova_design"] = "Mixed ANOVA"
    cfg["PCA_variables"] = [
        # Beam x Age
        "Nose x",
        "Nose y",
        "Ear base x",
        "Ear base y",
        "Front paw tao x",
        "Front paw tao y",
        "Wrist x",
        "Wrist y",
        "Elbow x",
        "Elbow y",
        "Lower Shoulder x",
        "Lower Shoulder y",
        "Upper Shoulder x",
        "Upper Shoulder y",
        "Iliac Crest x",
        "Iliac Crest y",
        "Hip x",
        "Hip y",
        "Knee x",
        "Knee y",
        "Ankle x",
        "Ankle y",
        "Hind paw tao x",
        "Hind paw tao y",
        "Tail base y",
        "Tail center x",
        "Tail center y",
        "Tail tip x",
        "Tail tip y",
        "Hind paw tao Velocity",
        "Hind paw tao Acceleration",
        "Ankle Velocity",
        "Ankle Acceleration",
        "Knee Velocity",
        "Knee Acceleration",
        "Hip Velocity",
        "Hip Acceleration",
        "Iliac Crest Velocity",
        "Iliac Crest Acceleration",
        "Ankle Angle",
        "Knee Angle",
        "Hip Angle",
        "Wrist Angle",
        "Elbow Angle",
        "Lower Shoulder Angle",
        "Iliac Crest Angle",
        "Ankle Angle Velocity",
        "Ankle Angle Acceleration",
        "Knee Angle Acceleration",
        "Hip Angle Velocity",
        "Hip Angle Acceleration",
        "Wrist Angle Velocity",
        "Wrist Angle Acceleration",
        "Elbow Angle Velocity",
        "Elbow Angle Acceleration",
        "Lower Shoulder Angle Velocity",
        "Lower Shoulder Angle Acceleration",
        "Iliac Crest Angle Velocity",
        "Iliac Crest Angle Acceleration",


        # PREVIOUS TESTING FEATURES
        # -----------------
        # "Hind paw tao y",
        # "Ankle y",
        # "Knee y",
        # "Ankle Angle",
        # "Knee Angle",
        # "Nose x",
        # "Knee y",
        # "Knee x",
        # "Knee Velocity",
        # "Knee Acceleration",
        # "Knee Angle",
        # "Knee Angle Velocity",
        # "Knee Angle Acceleration",
        # "Hip Angle",
    ]
    cfg["stats_variables"] = [
        # "Hind paw tao y",
        # "Ankle y",
        # "Knee y",
        # "Ankle Angle",
        # "Knee Angle",
        # "Nose x",
        # "Knee y",
        # "Knee x",
        # "Knee Velocity",
        # "Knee Acceleration",
        # "Knee Angle",
        # "Knee Angle Velocity",
        # "Knee Angle Acceleration",
        # "Hip Angle",
        # "Elbow Angle"


        # Beam x Age
        "Hip x",
        "Hip y",
        "Knee x",
        "Knee y",
        "Ankle x",
        "Ankle y",
        "Hind paw tao x",
        "Hind paw tao y",
        "Hind paw tao Velocity",
        "Ankle Velocity",
        "Knee Velocity",
        "Ankle Angle",
        "Knee Angle",
        "Hip Angle"
    ]
    # run
    autogaita.group(folderinfo, cfg)


# %% what happens if we just hit run
if __name__ == "__main__":
    group_dlcrun()



# %% imports
from autogaita.resources.utils import write_issues_to_textfile
import os
import shutil
import json
import pandas as pd
import numpy as np
import h5py

# %% constants
from autogaita.resources.constants import (
    TIME_COL,
    ISSUES_TXT_FILENAME,
    CONFIG_JSON_FILENAME,
)


# %% workflow step #1 - preparation


def some_prep(info, folderinfo, cfg):
    """Preparation of the data & cfg file for later analyses"""

    # ............................  unpack stuff  ......................................
    # => DON'T unpack (joint) cfg-keys that are tested later by check_and_expand_cfg
    # SLEAP-specific NOTE
    # => I commented out vars that we dont need but might need in the future
    name = info["name"]
    results_dir = info["results_dir"]
    data_string = folderinfo["data_string"]
    beam_string = folderinfo["beam_string"]
    sampling_rate = cfg["sampling_rate"]
    # IMPORTANT NOTE
    # => subtract_beam is hardcoded to False until I have data that allows me to test
    #    it properly (same for gait direction flipping @ end of this function)
    cfg["subtract_beam"] = False
    subtract_beam = cfg["subtract_beam"]
    convert_to_mm = cfg["convert_to_mm"]
    pixel_to_mm_ratio = cfg["pixel_to_mm_ratio"]
    standardise_y_at_SC_level = cfg["standardise_y_at_SC_level"]
    # invert_y_axis = cfg["invert_y_axis"]
    flip_gait_direction = cfg["flip_gait_direction"]
    analyse_average_x = cfg["analyse_average_x"]
    standardise_x_coordinates = cfg["standardise_x_coordinates"]
    standardise_y_to_a_joint = cfg["standardise_y_to_a_joint"]

    # .............................  move data  ........................................
    # => see if we can delete a previous runs results folder if existant. if not, it's a
    #    bit ugly since we only update results if filenames match...
    # => for example if angle acceleration not wanted in current run, but was stored in
    #    previous run, the previous run's figure is in the folder
    # => inform the user and leave this as is
    if os.path.exists(results_dir):
        try:
            shutil.rmtree(results_dir)
            move_data_to_folders(info, folderinfo)
        except OSError:
            move_data_to_folders(info, folderinfo)
            unable_to_rm_resdir_error = (
                "\n***********\n! WARNING !\n***********\n"
                + "Unable to remove previous Results subfolder of ID: "
                + name
                + "!\n Results will only be updated if filenames match!"
            )
            print(unable_to_rm_resdir_error)
            write_issues_to_textfile(unable_to_rm_resdir_error, info)
    else:
        move_data_to_folders(info, folderinfo)

    # .......  initialise Issues.txt & quick check for file existence  .................
    # Issues.txt - delete if saved in a previous run
    issues_txt_path = os.path.join(results_dir, ISSUES_TXT_FILENAME)
    if os.path.exists(issues_txt_path):
        os.remove(issues_txt_path)
    # read data & beam
    if not os.listdir(results_dir):
        no_files_error = (
            "\n******************\n! CRITICAL ERROR !\n******************\n"
            + "Unable to identify ANY RELEVANT FILES for "
            + name
        )
        write_issues_to_textfile(no_files_error, info)
        print(no_files_error)
        return

    # ............................  import data  .......................................
    # initialise dfs for user error handling
    datadf = pd.DataFrame(data=None)
    datadf_duplicate_error = ""
    beamdf = pd.DataFrame(data=None)
    beamdf_duplicate_error = ""
    # loop through folder and import data
    for filename in os.listdir(results_dir):
        if name + data_string + ".h5" in filename:
            if datadf.empty:
                datadf = h5_to_df(results_dir, filename)
            else:
                datadf_duplicate_error = (
                    "\n******************\n! CRITICAL ERROR !\n******************\n"
                    + "Multiple DATA .h5 files found for "
                    + name
                    + "!\nPlease make sure to only have one data file per ID."
                )
        if subtract_beam and name + beam_string + ".h5" in filename:
            if beamdf.empty:
                beamdf = h5_to_df(results_dir, filename)
            else:
                beamdf_duplicate_error = (
                    "\n******************\n! CRITICAL ERROR !\n******************\n"
                    + "Multiple BEAM .h5 files found for "
                    + name
                    + "!\nPlease make sure to only have one beam file per ID."
                )
    # handle errors now
    import_error_message = ""
    if datadf_duplicate_error:
        import_error_message += datadf_duplicate_error
    if datadf.empty:
        import_error_message += (
            "\n******************\n! CRITICAL ERROR !\n******************\n"
            + "No DATA .h5 file found for "
            + name
            + "!\nTry again!"
        )
    if subtract_beam:
        if beamdf_duplicate_error:
            import_error_message += beamdf_duplicate_error
        if beamdf.empty:
            import_error_message += (
                "\n******************\n! CRITICAL ERROR !\n******************\n"
                + "No BEAM .h5 file found for "
                + name
                + "!\nTry again!"
            )
    if import_error_message:
        write_issues_to_textfile(import_error_message, info)
        print(import_error_message)
        return  # make sure to stop execution if there is an issue!
    # create "data" as floats and depending on whether we subtracted beam or not
    if subtract_beam:
        data = pd.concat([datadf, beamdf], axis=1)
    else:
        data = datadf.copy(deep=True)
    data = data.astype(float)

    # ................  final data checks, conversions & additions  ....................
    # IMPORTANT - MAIN TESTS OF USER-INPUT VALIDITY OCCUR HERE!
    # => UNPACK VARS FROM CFG THAT ARE TESTED BY check_and_expand HERE, NOT EARLIER!
    cfg = check_and_expand_cfg(data, cfg, info)
    if cfg is None:  # some critical error occured
        return
    hind_joints = cfg["hind_joints"]
    fore_joints = cfg["fore_joints"]
    angles = cfg["angles"]
    beam_hind_jointadd = cfg["beam_hind_jointadd"]
    beam_fore_jointadd = cfg["beam_fore_jointadd"]
    direction_joint = cfg["direction_joint"]
    # important to unpack to vars hand not to cfg since cfg is overwritten in multiruns!
    x_standardisation_joint = cfg["x_standardisation_joint"][0]
    y_standardisation_joint = cfg["y_standardisation_joint"][0]
    # store config json file @ group path
    # !!! NU - do this @ mouse path!
    group_path = results_dir.split(name)[0]
    config_json_path = os.path.join(group_path, CONFIG_JSON_FILENAME)
    config_vars_to_json = {
        "sampling_rate": sampling_rate,
        "convert_to_mm": convert_to_mm,
        "standardise_y_at_SC_level": standardise_y_at_SC_level,
        "analyse_average_x": analyse_average_x,
        "standardise_x_coordinates": standardise_x_coordinates,
        "x_standardisation_joint": x_standardisation_joint,
        "standardise_y_to_a_joint": standardise_y_to_a_joint,
        "y_standardisation_joint": y_standardisation_joint,
        "hind_joints": hind_joints,
        "fore_joints": fore_joints,
        "angles": angles,
        "tracking_software": "SLEAP",
    }
    # note - using "w" will overwrite/truncate file, thus no need to remove it if exists
    with open(config_json_path, "w") as config_json_file:
        json.dump(config_vars_to_json, config_json_file, indent=4)
    # if we don't have a beam to subtract, standardise y to a joint's or global ymin = 0
    if not subtract_beam:
        y_min = float("inf")
        y_cols = [col for col in data.columns if col.endswith("y")]
        if standardise_y_to_a_joint:
            y_min = data[y_standardisation_joint + "y"].min()
        else:
            y_min = data[y_cols].min().min()
        data[y_cols] -= y_min
    # convert pixels to millimeters
    if convert_to_mm:
        for column in data.columns:
            # if might be unnecessary but I'm cautious as I don't know SLEAP data much
            if column.endswith("x") or column.endswith("y"):
                data[column] = data[column] / pixel_to_mm_ratio

    # IMPORTANT NOTE
    # => I keep gait direction flipping commented out until receiving data that allows
    #    me to test it properly
    # => Note that subtract_beam is hardcoded to False above for the same reason

    # check gait direction & DLC file validity
    # data = check_gait_direction(data, direction_joint, flip_gait_direction, info)
    data["Flipped"] = False  # because of IMPORTANT NOTE above

    # subtract the beam from the joints to standardise y
    # => bc. we simulate that all mice run from left to right, we can write:
    #     (note that we also flip beam x columns, but never y-columns!)
    # => & bc. we multiply y values by *-1 earlier, it's a neg_num - - neg_num
    #    pushing it towards zero.
    # => using list(set()) to ensure that we don't have duplicate values (if users
    #    should have provided them in both cfg vars by misstake)
    # => beam_col_left and right is provided by users
    if subtract_beam:
        # note beam_col_left/right are always lists in cfg!
        beam_col_left = cfg["beam_col_left"][0]
        beam_col_right = cfg["beam_col_right"][0]
        for joint in list(set(hind_joints + beam_hind_jointadd)):
            data[joint + "y"] = data[joint + "y"] - data[beam_col_left + "y"]
        for joint in list(set(fore_joints + beam_fore_jointadd)):
            data[joint + "y"] = data[joint + "y"] - data[beam_col_right + "y"]
        data.drop(columns=list(beamdf.columns), inplace=True)  # beam not needed anymore
    # add Time
    data[TIME_COL] = data.index * (1 / sampling_rate)
    # reorder the columns we added
    cols = [TIME_COL, "Flipped"]
    data = data[cols + [c for c in data.columns if c not in cols]]
    return data


# ..............................  helper functions  ....................................


def move_data_to_folders(info, folderinfo):
    """Copy data to new results_dir"""
    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    root_dir = folderinfo["root_dir"]
    data_string = folderinfo["data_string"]
    os.makedirs(results_dir)
    # move csv or h5 files
    for filename in os.listdir(root_dir):
        if name + data_string + ".csv" in filename:
            shutil.copy2(
                os.path.join(root_dir, filename),
                os.path.join(results_dir, filename),
            )
        elif name + data_string + ".h5" in filename:
            shutil.copy2(
                os.path.join(root_dir, filename),
                os.path.join(results_dir, filename),
            )


def h5_to_df(results_dir, filename):
    """Convert a SLEAP h5 file to a pandas dataframe used in gaita"""
    df = pd.DataFrame(data=None)
    with h5py.File(os.path.join(results_dir, filename), "r") as f:
        locations = f["tracks"][:].T
        node_names = [n.decode() for n in f["node_names"][:]]
        df.index = np.arange(np.shape(locations)[0])
        for node_idx, node_name in enumerate(node_names):
            for c, coord in enumerate(["x", "y"]):
                df[node_name + " " + coord] = locations[:, node_idx, c, 0]
    return df


def check_and_expand_cfg(data, cfg, info):
    """Test some important cfg variables and add new ones based on them

    Procedure
    ---------
    Check that no strings are empty
    Check that all strings end with a space character
    Check that all features are present in the dataset
    Add plot_joints & direction_joint
    Make sure to set dont_show_plots to True if Python is not in interactive mode
    If users subtract a beam, set normalise @ sc level to False
    String-checks for standardisation joints
    """

    # run the tests first
    # => note that beamcols & standardisation joints are tested separately further down.
    for cfg_key in [
        "angles",
        "hind_joints",
        "fore_joints",
    ]:
        cfg[cfg_key] = check_and_fix_cfg_strings(data, cfg, cfg_key, info)

    # add plot_joints - used for average plots & stick diagram
    hind_joints = cfg["hind_joints"]
    if cfg["plot_joint_number"] > len(hind_joints):
        fix_plot_joints_message = (
            "\n***********\n! WARNING !\n***********\n"
            + "You asked us to plot more hind limb joints than available!"
            + "\nNumber of joints to plot: "
            + str(cfg["plot_joint_number"])
            + "\nNumber of selected hindlimb joints: "
            + str(len(hind_joints))
            + "\n\nWe'll just plot the most we can :)"
        )
        write_issues_to_textfile(fix_plot_joints_message, info)
        print(fix_plot_joints_message)
        cfg["plot_joints"] = hind_joints
    else:
        cfg["plot_joints"] = hind_joints[: cfg["plot_joint_number"]]

    # check hindlimb joints & add direction_joint - used to determine gait direction
    if not hind_joints:  # no valid string after above cleaning
        no_hind_joint_message = (
            "\n******************\n! CRITICAL ERROR !\n******************\n"
            + "After testing your hind limb joint names, no valid joint was left to "
            + "perform gait direction checks on.\nPlease make sure that at least one "
            + "hind limb joint is provided & try again!"
        )
        write_issues_to_textfile(no_hind_joint_message, info)
        print(no_hind_joint_message)
        return
    cfg["direction_joint"] = hind_joints[0]

    # if subtracting beam, check identifier-strings & that beam colnames were valid.
    if cfg["subtract_beam"]:
        # first, let's check the strings
        # => note beamcols are lists of len=1 bc. of check function
        for cfg_key in [
            "beam_col_left",
            "beam_col_right",
            "beam_hind_jointadd",
            "beam_fore_jointadd",
        ]:
            cfg[cfg_key] = check_and_fix_cfg_strings(data, cfg, cfg_key, info)
        beam_col_error_message = (
            "\n******************\n! CRITICAL ERROR !\n******************\n"
            + "It seems like you want to standardise heights to a baseline (beam)."
            + "\nUnfortunately we were unable to find the y-columns you listed in "
            + "your beam's csv-file.\nPlease try again.\nInvalid beam side(s) was/were:"
        )
        beam_error = False  # check 3 possible cases
        if (cfg["beam_col_left"]) and (not cfg["beam_col_right"]):
            beam_error = True
            beam_col_error_message += "\n right beam!"
        elif (not cfg["beam_col_left"]) and (cfg["beam_col_right"]):
            beam_error = True
            beam_col_error_message += "\n left beam!"
        elif (not cfg["beam_col_left"]) and (not cfg["beam_col_right"]):
            beam_error = True
            beam_col_error_message += "\n both beams!"
        if beam_error:  # if any case was True, stop everything
            write_issues_to_textfile(beam_col_error_message, info)
            print(beam_col_error_message)
            return
    # never standardise @ SC level if user subtracted a beam
    if cfg["subtract_beam"] is True:
        cfg["standardise_y_at_SC_level"] = False
    # test x/y standardisation joints if needed
    broken_standardisation_joint = ""
    if cfg["standardise_x_coordinates"]:
        cfg["x_standardisation_joint"] = check_and_fix_cfg_strings(
            data, cfg, "x_standardisation_joint", info
        )
        if not cfg["x_standardisation_joint"]:
            broken_standardisation_joint += "x"
    if cfg["standardise_y_to_a_joint"]:
        cfg["y_standardisation_joint"] = check_and_fix_cfg_strings(
            data, cfg, "y_standardisation_joint", info
        )
        if not cfg["y_standardisation_joint"]:
            if broken_standardisation_joint:
                broken_standardisation_joint += " & y"
            else:
                broken_standardisation_joint += "y"
    if broken_standardisation_joint:
        no_standardisation_joint_message = (
            "\n******************\n! CRITICAL ERROR !\n******************\n"
            + "After testing your standardisation joints we found an issue with "
            + broken_standardisation_joint
            + "-coordinate standardisation joint."
            + "\n Cancelling AutoGaitA - please try again!"
        )
        write_issues_to_textfile(no_standardisation_joint_message, info)
        print(no_standardisation_joint_message)
        return

    return cfg


def check_and_fix_cfg_strings(data, cfg, cfg_key, info):
    """Check and fix strings in our joint & angle lists so that:
    1) They don't include empty strings
    2) All strings end with the space character
       => Important note: strings should never have the coordinate in them (since we do
          string + "y" for example throughout this code)
    3) All strings are valid columns of the DLC dataset
       => Note that x_standardisation_joint is tested against ending with "x" - rest
          against "y"
    """

    # work on this variable (we return to cfg[key] outside of here)
    string_variable = cfg[cfg_key]

    # easy checks: lists
    if type(string_variable) is list:
        string_variable = [s for s in string_variable if s]
        string_variable = [s if s.endswith(" ") else s + " " for s in string_variable]
        clean_string_list = []
        invalid_strings_message = ""
        for string in string_variable:
            if cfg_key == "x_standardisation_joint":
                if string + "x" in data.columns:
                    clean_string_list.append(string)
                else:
                    invalid_strings_message += "\n" + string
            else:  # for y_standardisation_joint (& all other cases)!!
                if string + "y" in data.columns:
                    clean_string_list.append(string)
                else:
                    invalid_strings_message += "\n" + string
        if invalid_strings_message:
            # print and save warning
            strings_warning = (
                "\n***********\n! WARNING !\n***********\n"
                + "Unable to find all "
                + cfg_key
                + " joints/key points you entered!"
                + "\n\nInvalid and thus removed were:"
                + invalid_strings_message
            )
            print(strings_warning)
            write_issues_to_textfile(strings_warning, info)
            # clean the string
            string_variable = clean_string_list
            # print & save info about remaining ("clean") strings
            clean_strings_message = "\nRemaining joints / key points are:"
            for string in clean_string_list:
                clean_strings_message += "\n" + string
            clean_strings_message += (
                "\n\nNote that capitalisation matters."
                + "\nIf you are running a batch analysis, we'll use this updated cfg "
                + "throughout\nCheck out the config.json file for the full cfg used."
            )
            print(clean_strings_message)
            write_issues_to_textfile(clean_strings_message, info)

    # things are more involved for angle dicts
    elif type(string_variable) is dict:
        # 1) test if the lists of all keys are equally long, if not throw out last idxs
        key_lengths = [len(string_variable[key]) for key in string_variable.keys()]
        if not all(key_length == key_lengths[0] for key_length in key_lengths):
            min_length = min(key_lengths)
            for key in string_variable:  # remove invalid idxs
                string_variable[key] = string_variable[key][:min_length]
            key_length_mismatch_message = (  # inform user
                "\n***********\n! WARNING !\n***********\n"
                + "\nLength-mismatch in angle configuration!"
                + "\nCheck angles' name/upper/lower-joint entries."
                + "\nOnly processing first "
                + str(min_length)
                + " entries."
            )
            print(key_length_mismatch_message)
            write_issues_to_textfile(key_length_mismatch_message, info)
        # 2) check if all strings in the angle dict are valid columns
        invalid_angletrio_message = ""
        invalid_idxs = []  # these idxs hold across the 3 keys of our angles-dict
        for key in string_variable:
            # important to do this first (else string + "y" is invalid)
            string_variable[key] = [
                s if s.endswith(" ") else s + " " for s in string_variable[key]
            ]
            # remove any empty or invalid (i.e. not in data) strings from our angletrio
            this_keys_missing_strings = ""
            for idx, string in enumerate(string_variable[key]):
                if string + "y" not in data.columns:  # checks for empty strings too
                    invalid_idxs.append(idx)
                    if not this_keys_missing_strings:  # first occurance
                        this_keys_missing_strings += "\nAngle's " + key + " key: "
                    this_keys_missing_strings += string + "(#" + str(idx + 1) + ") / "
            # string concat outside of idx-forloop above please
            invalid_angletrio_message += this_keys_missing_strings
        # if we have to remove idxs from all keys of our angles dict
        if invalid_angletrio_message:
            # print and save
            angles_warning = (
                "\n***********\n! WARNING !\n***********\n"
                + "Unable to find all "
                + cfg_key
                + " joints/key points you entered!"
                + "\n\nInvalid and thus removed were:"
                + invalid_angletrio_message
            )
            print(angles_warning)
            write_issues_to_textfile(angles_warning, info)
            # clean the dict
            invalid_idxs = list(set(invalid_idxs))  # remove duplicates
            for key in string_variable:
                string_variable[key] = [
                    string
                    for i, string in enumerate(string_variable[key])
                    if i not in invalid_idxs
                ]
            # print & save info about the remaining ("clean") angles dict
            clean_angles_message = "\nRemaining angle configuration is:"
            for i in range(len(string_variable["name"])):
                clean_angles_message += "\n\nAngle #" + str(i + 1)
                clean_angles_message += "\nName: " + string_variable["name"][i]
                clean_angles_message += (
                    "\nLower Joint: " + string_variable["lower_joint"][i]
                )
                clean_angles_message += (
                    "\nUpper Joint: " + string_variable["upper_joint"][i]
                )
            clean_angles_message += (
                "\n\nNote that capitalisation matters."
                + "\nIf you are running a batch analysis, we'll use this updated cfg "
                + "throughout\nCheck out the config.json file for the full cfg used."
            )
            print(clean_angles_message)
            write_issues_to_textfile(clean_angles_message, info)

    return string_variable


def flip_mouse_body(data, info):
    """If the mouse ran through the video frame from right to left simulate
    that it ran from left to right. For this just subtract all x-values of
    all x-columns from their respective maxima.
    ==> This preserves time-information important for SC extraction via table
    ==> This preserves y-information too
    ==> All analyses & plots are therefore comparable to mice that did really
        run from left to right
    """

    # 0) Tell the user that we are flipping their mouse
    message = (
        "\nDetermined gait direction of right => left - simulating it to be left => "
        + "right"
    )
    print(message)
    write_issues_to_textfile(message, info)

    # 1) Flip all rows in x columns only and subtract max from all vals
    flipped_data = data.copy()
    x_cols = [col for col in flipped_data.columns if col.endswith(" x")]
    for col in x_cols:
        flipped_data[col] = max(flipped_data[col]) - flipped_data[col]
    return flipped_data


def check_gait_direction(data, direction_joint, flip_gait_direction, info):
    """Check direction of gait - reverse it if needed

    Note
    ----
    Also using this check to check for DLC files being broken
    flip_gait_direction is only used after the check for DLC files being broken
    """

    data["Flipped"] = False
    enterframe = 0
    idx = 0
    flip_error_message = ""

    # beloow == True means that the mouse ran from right to left
    # => in this case we flip
    if np.median(data[direction_joint + "x"][: len(data) // 2]) > np.median(
        data[direction_joint + "x"][len(data) // 2 :]
    ):
        if flip_gait_direction:
            data = flip_mouse_body(data, info)
            data["Flipped"] = True
    return data



from autogaita.dlc.dlc_1_preparation import (
    some_prep,
    move_data_to_folders,
    check_and_expand_cfg,
    check_and_fix_cfg_strings,
    flip_mouse_body,
    check_gait_direction,
)
from hypothesis import given, strategies as st, settings, HealthCheck
import os
import copy
import math
import numpy as np
import pandas as pd
import pandas.testing as pdt
import pytest
import pdb


# %%................................  fixtures  ........................................
@pytest.fixture
def extract_data_using_some_prep(extract_info, extract_folderinfo, extract_cfg):
    data = some_prep(extract_info, extract_folderinfo, extract_cfg)
    return data


@pytest.fixture
def extract_info(tmp_path):
    info = {}
    info["mouse_num"] = 15
    info["run_num"] = 3
    info["name"] = "ID " + str(info["mouse_num"]) + " - Run " + str(info["run_num"])
    info["results_dir"] = os.path.join(tmp_path, info["name"])
    return info


@pytest.fixture
def extract_folderinfo():
    folderinfo = {}
    folderinfo["root_dir"] = "tests/test_data/dlc_data"
    folderinfo["sctable_filename"] = (
        "correct_annotation_table.xlsx"  # has to be an excel file
    )
    folderinfo["data_string"] = "SIMINewOct"
    folderinfo["beam_string"] = "BeamTraining"
    folderinfo["premouse_string"] = "Mouse"
    folderinfo["postmouse_string"] = "25mm"
    folderinfo["prerun_string"] = "run"
    folderinfo["postrun_string"] = "6DLC"
    return folderinfo


@pytest.fixture
def extract_cfg():
    cfg = {}
    cfg["sampling_rate"] = 100
    cfg["subtract_beam"] = True
    cfg["dont_show_plots"] = True
    cfg["convert_to_mm"] = True
    cfg["pixel_to_mm_ratio"] = 3.76
    cfg["x_sc_broken_threshold"] = 200
    cfg["y_sc_broken_threshold"] = 50
    cfg["x_acceleration"] = True
    cfg["angular_acceleration"] = True
    cfg["save_to_xls"] = True
    cfg["bin_num"] = 25
    cfg["plot_SE"] = True
    cfg["standardise_y_at_SC_level"] = False
    cfg["standardise_y_to_a_joint"] = True
    cfg["y_standardisation_joint"] = ["Knee"]
    cfg["plot_joint_number"] = 3
    cfg["color_palette"] = "viridis"
    cfg["legend_outside"] = True
    cfg["invert_y_axis"] = True
    cfg["flip_gait_direction"] = True
    cfg["analyse_average_x"] = True
    cfg["standardise_x_coordinates"] = True
    cfg["x_standardisation_joint"] = ["Hind paw tao"]
    cfg["hind_joints"] = ["Hind paw tao", "Ankle", "Knee", "Hip", "Iliac Crest"]
    cfg["fore_joints"] = [
        "Front paw tao ",
        "Wrist ",
        "Elbow ",
        "Lower Shoulder ",
        "Upper Shoulder ",
    ]
    cfg["beam_col_left"] = ["BeamLeft"]  # BEAM_COL_LEFT & _RIGHT must be lists of len=1
    cfg["beam_col_right"] = ["BeamRight"]
    cfg["beam_hind_jointadd"] = ["Tail base ", "Tail center ", "Tail tip "]
    cfg["beam_fore_jointadd"] = ["Nose ", "Ear base "]
    cfg["angles"] = {
        "name": ["Ankle ", "Knee ", "Hip "],
        "lower_joint": ["Hind paw tao ", "Ankle ", "Knee "],
        "upper_joint": ["Knee ", "Hip ", "Iliac Crest "],
    }
    return cfg


# %%..............................  preparation  .......................................
#                       AN IMPORTANT NOTE ON THESE UNIT TESTS!
# Calling check_and_expand cfg outside of some_prep with the
# export_data_using_some_prep fixture leads to the cfg that is returned by
# check_and_expand to be None since the data var DOES NOT INCLUDE the beam!
# => We thus set cfg["subtract_beam"] to False prior to calling it (see e.g.the
#    plot_joint test)


def test_wrong_data_and_beam_strings(extract_info, extract_folderinfo, extract_cfg):
    extract_folderinfo["beam_string"] = extract_folderinfo["data_string"]
    some_prep(extract_info, extract_folderinfo, extract_cfg)
    with open(os.path.join(extract_info["results_dir"], "Issues.txt")) as f:
        content = f.read()
    assert "Your data & baseline (beam) identifiers ([G] in our" in content


def test_wrong_postmouse_string(extract_info, extract_folderinfo, extract_cfg):
    extract_folderinfo["postmouse_string"] = "this_is_a_test"
    some_prep(extract_info, extract_folderinfo, extract_cfg)
    with open(os.path.join(extract_info["results_dir"], "Issues.txt")) as f:
        content = f.read()
    assert "Unable to identify ANY RELEVANT FILES for" in content


def test_global_min_standardisation(extract_info, extract_folderinfo, extract_cfg):
    extract_cfg["subtract_beam"] = False
    extract_cfg["standardise_y_to_a_joint"] = False
    data = some_prep(extract_info, extract_folderinfo, extract_cfg)
    y_cols = [c for c in data.columns if c.endswith(" y")]
    assert data[y_cols].min().min() == 0
    # approach here is find difference between global & standardisation joint minma and
    # see if all y cols' difference is equal to that
    # => this implies that joint-based y-standardisation worked
    global_and_standardisation_joints_y_min_diff = data[
        extract_cfg["y_standardisation_joint"][0] + " y"
    ].min()
    global_min_data = data.copy()
    extract_cfg["standardise_y_to_a_joint"] = True
    data = some_prep(extract_info, extract_folderinfo, extract_cfg)
    assert np.allclose(  # use np.allclose here because we are comparing arrays
        global_min_data[y_cols],
        data[y_cols] + global_and_standardisation_joints_y_min_diff,
        atol=1e-9,
    )


def test_datas_indexing_and_time_column(extract_info, extract_folderinfo, extract_cfg):
    for extract_cfg["sampling_rate"] in [50, 500, 5000]:
        data = some_prep(extract_info, extract_folderinfo, extract_cfg)
        # use isclose here because there are some floating point things going on (eg. 1.
        # 66 and 1.660 for sampling rate of 500)
        assert math.isclose(
            data["Time"].max(),
            (len(data) - 1) / (1 * extract_cfg["sampling_rate"]),
            rel_tol=1e-9,
        )


def test_cols_we_added_to_data(extract_info, extract_folderinfo, extract_cfg):
    data = some_prep(extract_info, extract_folderinfo, extract_cfg)
    assert (data.columns[0] == "Time") & (data.columns[1] == "Flipped")


def test_move_csv_datafile_to_results_dir(extract_info, extract_folderinfo):
    move_data_to_folders(extract_info, extract_folderinfo)
    for file in os.listdir(extract_info["results_dir"]):
        if file.endswith(".csv"):
            assert (
                (extract_folderinfo["premouse_string"] in file)
                & (extract_folderinfo["postmouse_string"] in file)
                & (extract_folderinfo["prerun_string"] in file)
                & (extract_folderinfo["postrun_string"] in file)
            )


def test_error_if_no_cfgkey_joints(extract_info, extract_folderinfo, extract_cfg):
    full_cfg = copy.deepcopy(extract_cfg)  # no referencing here - we need copies!
    for cfg_key in [
        "hind_joints",
        "x_standardisation_joint",
        "y_standardisation_joint",
    ]:
        extract_cfg = copy.deepcopy(full_cfg)  # here too!
        extract_cfg[cfg_key] = ["not_in_data"]
        data = some_prep(extract_info, extract_folderinfo, extract_cfg)
        with open(os.path.join(extract_info["results_dir"], "Issues.txt")) as f:
            content = f.read()
        if cfg_key == "hind_joints":
            assert "hind limb joint names" in content
        elif cfg_key == "x_standardisation_joint":
            assert "x-coordinate standardisation joint" in content
        elif cfg_key == "y_standardisation_joint":
            assert "y-coordinate standardisation joint" in content
        assert data is None
    # cannot loop case of x & y joints being broken
    extract_cfg = copy.deepcopy(full_cfg)
    extract_cfg["x_standardisation_joint"] = ["not_in_data"]
    extract_cfg["y_standardisation_joint"] = ["not_in_data"]
    data = some_prep(extract_info, extract_folderinfo, extract_cfg)
    with open(os.path.join(extract_info["results_dir"], "Issues.txt")) as f:
        content = f.read()
        assert "x & y-coordinate standardisation joint" in content
        assert data is None


def test_plot_joint_error(extract_data_using_some_prep, extract_cfg, extract_info):
    extract_cfg["plot_joint_number"] = 2000
    extract_cfg["subtract_beam"] = False
    check_and_expand_cfg(extract_data_using_some_prep, extract_cfg, extract_info)
    with open(os.path.join(extract_info["results_dir"], "Issues.txt")) as f:
        content = f.read()
    assert "we can :)" in content
    extract_cfg = check_and_expand_cfg(
        extract_data_using_some_prep, extract_cfg, extract_info
    )
    assert extract_cfg["plot_joints"] == extract_cfg["hind_joints"]
    extract_cfg["plot_joint_number"] = 2
    extract_cfg = check_and_expand_cfg(
        extract_data_using_some_prep, extract_cfg, extract_info
    )
    assert extract_cfg["plot_joints"] == extract_cfg["hind_joints"][:2]


@given(test_list=st.lists(st.text(), min_size=1))
@settings(suppress_health_check=[HealthCheck.function_scoped_fixture])
def test_removal_of_wrong_strings_from_cfg_key(
    test_list, extract_data_using_some_prep, extract_cfg, extract_info
):
    cfg_key = "hind_joints"  # irrelevant since property testing
    extract_cfg[cfg_key] = test_list
    test_result = check_and_fix_cfg_strings(
        extract_data_using_some_prep, extract_cfg, cfg_key, extract_info
    )
    assert not test_result  # empty list is falsey


def test_flip_mouse_body(extract_info, extract_folderinfo, extract_cfg):
    extract_cfg["flip_gait_direction"] = False
    test_data = some_prep(extract_info, extract_folderinfo, extract_cfg)
    test_flipped_data = test_data.copy()
    print("flipped_data pre flipping:")
    print(test_flipped_data)
    test_flipped_data = flip_mouse_body(test_data, extract_info)
    # flipped_data = data.copy()
    # test_data = data.copy()
    print("flipped_data post flipping:")
    print(test_flipped_data)
    for col in test_flipped_data.columns:
        if col.endswith("x"):
            flipped_test_series = test_flipped_data[col]
            print(f"Processing column: {col}")
            print("flipped_test_series:")
            print(flipped_test_series)
            # pytest.set_trace()
            test_series = max(test_data[col]) - test_data[col]
            print("test_series:")
            print(test_series)
            flipped_test_series = flipped_test_series.astype(float)
            print("flipped_test_series after conversion:")
            print(flipped_test_series)
            pdt.assert_series_equal(test_series, flipped_test_series)


def test_check_gait_direction(extract_data_using_some_prep, extract_cfg, extract_info):
    direction_joint = extract_cfg["direction_joint"]
    flip_gait_direction = True  # 1) test broken DLC data (empty)
    broken_data = pd.DataFrame(data=None, columns=extract_data_using_some_prep.columns)
    check_gait_direction(
        broken_data, direction_joint, flip_gait_direction, extract_info
    )
    with open(os.path.join(extract_info["results_dir"], "Issues.txt")) as f:
        content = f.read()
    assert "Unable to determine gait direction!" in content
    flip_this_data = pd.DataFrame(
        data=None, columns=extract_data_using_some_prep.columns
    )  # 2) test data that has to be flipped
    data_len = 20
    flip_this_data[direction_joint + "likelihood"] = [0.99] * data_len
    first_half = [10] * int((data_len / 2))
    second_half = [-10] * int((data_len / 2))
    x_coords = first_half + second_half
    flip_this_data[direction_joint + "x"] = x_coords
    flip_this_data = check_gait_direction(
        flip_this_data, direction_joint, flip_gait_direction, extract_info
    )
    assert flip_this_data["Flipped"][0] == True



# %% imports
from autogaita.resources.utils import write_issues_to_textfile
from autogaita.common2D.common2D_utils import (
    check_cycle_out_of_bounds,
    check_cycle_duplicates,
    check_cycle_order,
    check_tracking,
    handle_issues,
)
import os
import pandas as pd
import numpy as np

# %% constants
from autogaita.common2D.common2D_constants import (
    SCXLS_MOUSECOLS,
    SCXLS_SCCOLS,
    SWINGSTART_COL,
    STANCEEND_COL,
)

# %% workflow step #2 - SC extraction (reading user-provided SC Table)


def extract_stepcycles(data, info, folderinfo, cfg):
    """Read XLS file with SC annotations, find correct row & return all_cycles"""

    # ...............................  preparation  ....................................
    # unpack
    name = info["name"]
    root_dir = folderinfo["root_dir"]
    sctable_filename = folderinfo["sctable_filename"]
    sampling_rate = cfg["sampling_rate"]

    # check if excel file is .xlsx or .xls, if none found try to fix it
    if (".xls" in sctable_filename) | (".xlsx" in sctable_filename):
        if os.path.exists(os.path.join(root_dir, sctable_filename)):
            SCdf = pd.read_excel(os.path.join(root_dir, sctable_filename))
        else:
            raise FileNotFoundError(
                "No Annotation Table found! sctable_filename has to be @ root_dir"
            )
    else:
        # in cases below use string-concat (+) - otherwise xls added as path
        if os.path.exists(os.path.join(root_dir, sctable_filename + ".xls")):
            SCdf = pd.read_excel(os.path.join(root_dir, sctable_filename + ".xls"))
        elif os.path.exists(os.path.join(root_dir, sctable_filename + ".xlsx")):
            SCdf = pd.read_excel(os.path.join(root_dir, sctable_filename + ".xlsx"))
        else:
            raise FileNotFoundError(
                "No Annotation Table found! sctable_filename has to be @ root_dir"
            )
    # see if table columns are labelled correctly (try a couple to allow user typos)
    valid_col_flags = [False, False]
    header_columns = ["", ""]
    for h, header in enumerate([SCXLS_MOUSECOLS, SCXLS_SCCOLS]):
        for header_col in header:
            if header_col in SCdf.columns:
                valid_col_flags[h] = True
                header_columns[h] = header_col
                break
    if not all(valid_col_flags):
        handle_issues("wrong_scxls_colnames", info)
        return
    # find our info columns & rows
    mouse_col = SCdf.columns.get_loc(header_columns[0])  # INDEXING! (see list above)
    sc_col = SCdf.columns.get_loc(header_columns[1])
    SCdf[header_columns[0]] = SCdf[header_columns[0]].astype(str)
    mouse_row = SCdf.index[SCdf[header_columns[0]] == name]  # find this mouse
    # this mouse was not included in sc xls
    if len(mouse_row) == 0:
        handle_issues("no_mouse", info)
        return
    # this mouse was included more than once
    if len(mouse_row) > 1:
        handle_issues("double_mouse", info)
        return

    # next_mouse_idx = mouse_row  # search idx of first row of next mouse

    # ..............................  main xls read  ...................................
    sc_num = 0
    for column in SCdf.columns:
        if STANCEEND_COL in column:
            if np.isnan(SCdf[column][mouse_row].values[0]) == False:
                sc_num += 1
    if sc_num == 0:
        handle_issues("no_scs", info)
        return
    user_scnum = SCdf.iloc[mouse_row, sc_col].values[0]  # sanity check input
    if user_scnum != sc_num:  # warn the user, take the values we found
        this_message = (
            "\n***********\n! WARNING !\n***********\n"
            + "Mismatch between stepcycle number of SC Number column & "
            + "entries in swing/stance latency columns!"
            + "\nUsing all valid swing/stance entries."
        )
        print(this_message)
        write_issues_to_textfile(this_message, info)

    # ...........................  idxs to all_cycles  .................................
    # use value we found, loop over all runs, throw all scs into all_cycles
    all_cycles = [[None, None] for s in range(sc_num)]  # fill :sc_num x 2 list
    for s in range(sc_num):
        if s == 0:
            start_col = SCdf.columns.get_loc(SWINGSTART_COL)
            end_col = SCdf.columns.get_loc(STANCEEND_COL)
        else:
            # str(s) because colnames match s for s>0!
            start_col = SCdf.columns.get_loc(SWINGSTART_COL + "." + str(s))
            end_col = SCdf.columns.get_loc(STANCEEND_COL + "." + str(s))
        user_scnum += 1
        # extract the SC times
        start_in_s = float(SCdf.iloc[mouse_row, start_col].values[0])
        end_in_s = float(SCdf.iloc[mouse_row, end_col].values[0])
        # see if we are rounding to fix inaccurate user input
        # => account for python's float precision leading to inaccuracies
        # => two important steps here (sanity_check_vals only used for these checks)
        # 1. round to 10th decimal to fix python making
        #    3211.999999999999999995 out of 3212
        sanity_check_start = round(start_in_s * sampling_rate, 10)
        sanity_check_end = round(end_in_s * sampling_rate, 10)
        # 2. comparing abs(sanity check vals) to 1e-7 just to be 1000% sure
        if (abs(sanity_check_start % 1) > 1e-7) | (abs(sanity_check_end % 1) > 1e-7):
            round_message = (
                "\n***********\n! WARNING !\n***********\n"
                + "SC latencies of "
                + str(start_in_s)
                + "s to "
                + str(end_in_s)
                + "s were not provided in units of the frame rate!"
                + "\nWe thus use the previous possible frame(s)."
                + "\nDouble check if this worked as expected or fix annotation table!"
            )
            print(round_message)
            write_issues_to_textfile(round_message, info)
        # assign to all_cycles (note int() rounds down!)
        all_cycles[s][0] = int(start_in_s * sampling_rate)
        all_cycles[s][1] = int(end_in_s * sampling_rate)
        # check if we are in data-bounds
        if (all_cycles[s][0] in data.index) & (all_cycles[s][1] in data.index):
            pass
        else:
            all_cycles[s] = [None, None]  # so they can be cleaned later
            this_message = (
                "\n***********\n! WARNING !\n***********\n"
                + "SC latencies of: "
                + str(start_in_s)
                + "s to "
                + str(end_in_s)
                + "s not in data/video range!"
                + "\nSkipping!"
            )
            print(this_message)
            write_issues_to_textfile(this_message, info)

    # ............................  clean all_cycles  ..................................
    # check if we skipped latencies because they were out of data-bounds
    all_cycles = check_cycle_out_of_bounds(all_cycles)
    if all_cycles:  # can be None if all SCs were out of bounds
        # check if there are any duplicates (e.g., SC2's start-lat == SC1's end-lat)
        all_cycles = check_cycle_duplicates(all_cycles)
        # check if user input progressively later latencies
        all_cycles = check_cycle_order(all_cycles, info)
        # check if tracking broke for any SCs - if so remove them
        all_cycles = check_tracking(data, info, all_cycles, cfg)
    return all_cycles



from autogaita.resources.constants import TIME_COL

# LEG stuff
LEGS = ["left", "right"]
LEGS_COLFORMAT = [", left ", ", right "]
OUTPUTS = LEGS + ["both"]
# sc extraction
SCXLS_SUBJCOLS = [
    "Participant",
    "participant",
    "Animal",
    "animal",
    "Subject",
    "subject",
    "ID",
    "id",
]  # SC XLS info
SCXLS_LEGCOLS = ["Leg", "leg", "Legs", "legs", "Side", "side"]
SCXLS_RUNCOLS = ["Run", "run", "Runs", "runs", "Trial", "trial", "Trials", "trials"]
SCXLS_SCCOLS = ["SC Number", "SC number", "sc number", "SC Num", "sc num", "SC num"]
SWINGSTART_COL = "Swing (ti)"
STANCEEND_COL = "Stance (te)"
# simulate walking direction being left to right
SEARCH_WIN_TURN_TIME = 500  # 5 seconds
# export results as xlsx
ORIGINAL_XLS_FILENAME = " - Original Stepcycles"  # filenames of sheet exports
NORMALISED_XLS_FILENAME = " - Normalised Stepcycles"
AVERAGE_XLS_FILENAME = " - Average Stepcycle"
STD_XLS_FILENAME = " - Standard Devs. Stepcycle"
SEPARATOR_IDX = 1  # idx of dfs whenever we have separator rows
LEG_COL = "Leg"
EXCLUDED_COLS_IN_AV_STD_DFS = [TIME_COL, LEG_COL]
REORDER_COLS_IN_STEP_NORMDATA = [TIME_COL, LEG_COL]
# plot stuff
SC_LAT_LEGEND_FONTSIZE = 6
ANGLE_PLOTS_YLIMITS = [80, 190]
STICK_LINEWIDTH = 0.5
# Plot GUI colors
FG_COLOR = "#c0737a"  # dusty rose
HOVER_COLOR = "#b5485d"  # dark rose



# main gui
from .gui.main_gui import run_gui  # autogaita.run_gui()

# 3 sub-guis
from .gui.dlc_gui import run_dlc_gui  # autogaita.run_dlc_gui()
from .gui.sleap_gui import run_sleap_gui  # autogaita.run_sleap_gui()
from .gui.universal3D_gui import (
    run_universal3D_gui,
)  # autogaita.run_universal3D_gui()
from .gui.group_gui import run_group_gui  # autogaita.run_group_gui()

# 4 main functions
from .dlc.dlc_main import dlc  # autogaita.dlc(info, folderinfo, cfg)
from .sleap.sleap_main import sleap  # autogaita.sleap(info, folderinfo, cfg)
from .universal3D.universal3D_main import (
    universal3D,
)  # autogaita.universal3D(info, folderinfo, cfg)
from .group.group_main import group  # autogaita.group(folderinfo, cfg)

# 6 batchrun functions - call via e.g. autogaita.dlc_singlerun()
from .batchrun_scripts.dlc_singlerun import dlc_singlerun
from .batchrun_scripts.dlc_multirun import dlc_multirun
from .batchrun_scripts.sleap_singlerun import sleap_singlerun
from .batchrun_scripts.universal3D_multirun import universal3D_multirun
from .batchrun_scripts.universal3D_singlerun import universal3D_singlerun
from .batchrun_scripts.group_dlcrun import group_dlcrun
from .batchrun_scripts.group_universal3Drun import group_universal3Drun



# %% imports
import os
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

# %% constants
from autogaita.resources.constants import TIME_COL
from autogaita.common2D.common2D_constants import (
    SC_LAT_LEGEND_FONTSIZE,
)


# %% workflow step #4 - extract SCs from all_steps_data (# !!! NU - load XLS) and plot


# ..............................  master function  .............................

# A note on x-standardisation (14.08.2024)
# => If x-standardisation was performed, original as well as x-standardised dfs are
#    generated and exported to xls
# => Our plotting functions only used all_steps_data, average_data & std_data
# => Conveniently, if x-standardisation is performed, all_steps_data DOES NOT include
#    x-standardisation and is thus still the correct choice for the step-cycle level
#    plots (#1-#5)
# => On the other hand, average_data & std_data (plots #6-12) DO and SHOULD include
#    x-standardisation
# => Note that there is a slight discrepancy because angle-plots are based on
#    x-standardised values  for the step-cycle level plots (#3) but not for the average
#    plots (#8)
#   -- This is not an issue since x-standardisation does not affect angles, even though
#      the values of x change because they change by a constant for all joints (there
#      is a unit test for this)

# A note on updated colour cyclers after pull request that was merged 20.06.2024
# => Using color palettes instead of colour maps as I had previously means that
#    we cycle through neighbouring colours
# => I initially implemented an "equally distant" approach.
# => So for viridis and n=2 (e.g. if 2 groups) it would be purple and yellow
#    (as far away as possible)
# => Now it is dark blue and green
# => Updated approach is aesthetically more pleasing IMO.
# => However it does have the risk of not being able to tell the colours in
#    some cases - e.g. if some accelerations are very overlapping.
# => But - because users can in theses cases just choose a colour palette that
#    in itself has categorical colours (Set1, Dark2, etc.) I still keep the new
#    behaviour
# => Nonetheless, in case you want to use the "old behaviour" at some point it
#    would be coded as commented out in plot_joint_y_by_x


def plot_results(info, results, cfg, plot_panel_instance):
    """Plot results - y coords by x coords & average angles over SC %"""
    # unpack
    fore_joints = cfg["fore_joints"]
    angles = cfg["angles"]
    all_steps_data = results["all_steps_data"]
    average_data = results["average_data"]
    std_data = results["std_data"]
    x_acceleration = cfg["x_acceleration"]
    angular_acceleration = cfg["angular_acceleration"]
    analyse_average_x = cfg["analyse_average_x"]
    dont_show_plots = cfg["dont_show_plots"]
    if dont_show_plots:
        plt.switch_backend("Agg")

    # ....................0 - extract SCs from all_steps_data...........................
    sc_idxs = extract_sc_idxs(all_steps_data)
    cfg["sc_num"] = len(sc_idxs)  # add number of scs for plotting SE if wanted

    # .........................1 - y coords by x coords.................................
    plot_joint_y_by_x(all_steps_data, sc_idxs, info, cfg, plot_panel_instance)

    # .....................2 - x coords by time (optional!).............................
    if analyse_average_x:
        plot_x_by_time(all_steps_data, sc_idxs, info, cfg, plot_panel_instance)

    # ...............................3 - angles by time.................................
    if angles["name"]:
        plot_angles_by_time(all_steps_data, sc_idxs, info, cfg, plot_panel_instance)
    # regularly closing figures to save memory
    # => no problem to do this since we pass figure-vars to save-functions and PlotPanel
    plt.close("all")

    # ..........................4 - hindlimb stick diagram..............................
    plot_hindlimb_stickdiagram(all_steps_data, sc_idxs, info, cfg, plot_panel_instance)

    # ...........................5 - forelimb stick diagram.............................
    if fore_joints:
        plot_forelimb_stickdiagram(
            all_steps_data, sc_idxs, info, cfg, plot_panel_instance
        )

    # .....................6 - average joints' y over SC percentage.....................
    plot_joint_y_by_average_SC(average_data, std_data, info, cfg, plot_panel_instance)

    # ...............7 - average joints' x over SC percentage (optional!)...............
    if analyse_average_x:
        plot_joint_x_by_average_SC(
            average_data, std_data, info, cfg, plot_panel_instance
        )
    plt.close("all")

    # ........................8 - average angles over SC percentage.....................
    if angles["name"]:
        plot_angles_by_average_SC(
            average_data, std_data, info, cfg, plot_panel_instance
        )

    # .................9 - average x velocities over SC percentage......................
    plot_x_velocities_by_average_SC(
        average_data, std_data, info, cfg, plot_panel_instance
    )

    # .............10 - average angular velocities over SC percentage...................
    if angles["name"]:
        plot_angular_velocities_by_average_SC(
            average_data, std_data, info, cfg, plot_panel_instance
        )

    # ............optional - 11 - average x acceleration over SC percentage.............
    if x_acceleration:
        plot_x_acceleration_by_average_SC(
            average_data, std_data, info, cfg, plot_panel_instance
        )

    # .........optional - 12 - average angular acceleration over SC percentage..........
    if angles["name"]:
        if angular_acceleration:
            plot_angular_acceleration_by_average_SC(
                average_data, std_data, info, cfg, plot_panel_instance
            )
    plt.close("all")

    # ........................optional - 13 - build plot panel..........................
    if dont_show_plots is True:
        pass  # going on without building the plot window
    elif dont_show_plots is False:  # -> show plot panel
        # Destroy loading screen and build plot panel with all figures
        plot_panel_instance.destroy_plot_panel_loading_screen()
        plot_panel_instance.build_plot_panel()


# ..................................  inner functions  .................................


def extract_sc_idxs(all_steps_data):
    """0 - Prepare stepcycles on original-length (non-normalised) SCs"""
    # A note on all_steps_data & nan (see xls_separations):
    # ==> (Using range & iloc so we don't have to subtract 1 to nan-idxs)
    # ==> if there is more than 1 SC found, the first row of nan indicates the
    #     END of SC 1
    # ==> the last row of nan indicates the START of the last SC
    # ==> everything inbetween is alternatingly: (if you add 1 to nan-idx) the
    #     start of an SC + (if you subtract -1 to nan-idx) the end of that SC
    # ==> E.g.: separations[1]+1 is 1st idx of SC2 - separations[2]-1 is last
    #     idx of SC2
    check_col = all_steps_data.columns[0]  # take the first col to find nan-separators
    xls_separations = np.where(pd.isnull(all_steps_data[check_col]))[0]
    sc_idxs = []
    # the next line means that we have exactly one step, because we would not
    # build all_steps_data (and have results in the first place) if there was no step
    # Thus, if xls_sep. is empty (len=0) it means that no separations were
    # there, i.e., 1 SC
    if len(xls_separations) == 0:
        sc_idxs.append(range(0, len(all_steps_data)))  # I can do this bc. only 1 SC
    else:
        for b in range(len(xls_separations)):
            if b == 0:
                # SC1 - 0 to (not including) nan/end-idx
                sc_idxs.append(range(xls_separations[b]))
            elif b > 0:  # inbetween SCs
                if (b % 2) == 0:
                    sc_idxs.append(
                        range(
                            xls_separations[b - 1] + 1,  # add 1=start
                            xls_separations[b],
                        )
                    )
            # last SC - I can write it this way because b will always be odd if
            # it refers to the start of a stepcycle & thus: possibly the last
            if xls_separations[b] == xls_separations[-1]:
                sc_idxs.append(range(xls_separations[-1] + 1, len(all_steps_data)))
    return sc_idxs


def plot_joint_y_by_x(all_steps_data, sc_idxs, info, cfg, plot_panel_instance):
    """1 - Plot joints' y coordinates as a function of their x for each SC"""

    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    dont_show_plots = cfg["dont_show_plots"]
    convert_to_mm = cfg["convert_to_mm"]
    plot_joints = cfg["plot_joints"]
    sampling_rate = cfg["sampling_rate"]
    legend_outside = cfg["legend_outside"]
    color_palette = cfg["color_palette"]

    # some prep
    sc_num = len(sc_idxs)
    f = [[] for _ in range(len(plot_joints))]
    ax = [[] for _ in range(len(plot_joints))]

    # plot
    for j, joint in enumerate(plot_joints):  # joint loop (figures)
        f[j], ax[j] = plt.subplots(1, 1)
        # What "Old" colormap approach would look like with seaborn
        # this_map = sns.color_palette(cfg["color_palette"], as_cmap=True)
        # ax[j].set_prop_cycle(plt.cycler("color", this_map(np.linspace(0, 1, sc_num))))
        ax[j].set_prop_cycle(  # New color palette approach
            plt.cycler("color", sns.color_palette(color_palette, sc_num))
        )
        ax[j].set_title(name + " - " + joint + "Y")
        x_col_idx = all_steps_data.columns.get_loc(joint + "x")
        y_col_idx = all_steps_data.columns.get_loc(joint + "y")
        time_col_idx = all_steps_data.columns.get_loc(TIME_COL)
        for s in range(sc_num):
            this_x = all_steps_data.iloc[sc_idxs[s], x_col_idx]
            this_y = all_steps_data.iloc[sc_idxs[s], y_col_idx]
            this_label = generate_sc_latency_label(
                all_steps_data, sc_idxs[s], sampling_rate, time_col_idx
            )
            ax[j].plot(this_x, this_y, label=this_label)
        # legend adjustments
        if legend_outside is True:
            ax[j].legend(
                fontsize=SC_LAT_LEGEND_FONTSIZE,
                loc="center left",
                bbox_to_anchor=(1, 0.5),
            )
        elif legend_outside is False:
            ax[j].legend(fontsize=SC_LAT_LEGEND_FONTSIZE)
        # labels & conversion
        ax[j].set_xlabel("x")  # will be used by conversion func if we convert
        ax[j].set_ylabel("y")
        if convert_to_mm:
            tickconvert_mm_to_cm(ax[j], "both")
        else:
            ax[j].set_xlabel("x (pixels)")
            ax[j].set_ylabel("y (pixels)")
        figure_file_string = " - " + joint + "Y by X coordinates"
        save_figures(f[j], results_dir, name, figure_file_string)
        if dont_show_plots:
            plt.close(f[j])

        # add figure to plot panel figures list
        if dont_show_plots is False:  # -> show plot panel
            plot_panel_instance.figures.append(f[j])


def plot_x_by_time(all_steps_data, sc_idxs, info, cfg, plot_panel_instance):
    """2 - Plot joints' x coordinates as a function of time for each SC (optional)"""

    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    dont_show_plots = cfg["dont_show_plots"]
    convert_to_mm = cfg["convert_to_mm"]
    plot_joints = cfg["plot_joints"]
    sampling_rate = cfg["sampling_rate"]
    legend_outside = cfg["legend_outside"]
    color_palette = cfg["color_palette"]

    # some prep
    sc_num = len(sc_idxs)
    f = [[] for _ in range(len(plot_joints))]
    ax = [[] for _ in range(len(plot_joints))]

    # plot
    for j, joint in enumerate(plot_joints):  # joint loop (figures)
        f[j], ax[j] = plt.subplots(1, 1)
        ax[j].set_prop_cycle(
            plt.cycler("color", sns.color_palette(color_palette, sc_num))
        )
        ax[j].set_title(name + " - " + joint + "X")
        time_col_idx = all_steps_data.columns.get_loc(TIME_COL)
        x_col_idx = all_steps_data.columns.get_loc(joint + "x")
        for s in range(sc_num):
            this_x = all_steps_data.iloc[sc_idxs[s], time_col_idx]
            this_y = all_steps_data.iloc[sc_idxs[s], x_col_idx]
            this_label = generate_sc_latency_label(
                all_steps_data, sc_idxs[s], sampling_rate, time_col_idx
            )
            ax[j].plot(this_x, this_y, label=this_label)
        # legend adjustments
        if legend_outside is True:
            ax[j].legend(
                fontsize=SC_LAT_LEGEND_FONTSIZE,
                loc="center left",
                bbox_to_anchor=(1, 0.5),
            )
        elif legend_outside is False:
            ax[j].legend(fontsize=SC_LAT_LEGEND_FONTSIZE)
        # labels & conversion
        ax[j].set_xlabel("Time (s)")  # will be used by conversion func if we convert
        ax[j].set_ylabel("x")
        if convert_to_mm:
            tickconvert_mm_to_cm(ax[j], "y")
        else:
            ax[j].set_ylabel("x (pixels)")
        figure_file_string = " - " + joint + "X coordinate by Time"
        save_figures(f[j], results_dir, name, figure_file_string)
        if dont_show_plots:
            plt.close(f[j])

        # add figure to plot panel figures list
        if dont_show_plots is False:  # -> show plot panel
            plot_panel_instance.figures.append(f[j])


def plot_angles_by_time(all_steps_data, sc_idxs, info, cfg, plot_panel_instance):
    """3 - Plot joints' angles as a function of time for each SC"""

    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    dont_show_plots = cfg["dont_show_plots"]
    angles = cfg["angles"]
    sampling_rate = cfg["sampling_rate"]
    legend_outside = cfg["legend_outside"]
    color_palette = cfg["color_palette"]

    # some prep
    sc_num = len(sc_idxs)
    f = [[] for _ in range(len(angles["name"]))]
    ax = [[] for _ in range(len(angles["name"]))]

    # plot
    for a, angle in enumerate(angles["name"]):  # angle loop (figures)
        f[a], ax[a] = plt.subplots(1, 1)
        ax[a].set_prop_cycle(
            plt.cycler("color", sns.color_palette(color_palette, sc_num))
        )
        ax[a].set_title(name + " - " + angle + "Angle")
        ax[a].set_ylabel("Angle")
        ax[a].set_xlabel("Time (s)")
        x_col_idx = all_steps_data.columns.get_loc(TIME_COL)
        y_col_idx = all_steps_data.columns.get_loc(angle + "Angle")
        time_col_idx = all_steps_data.columns.get_loc(TIME_COL)
        for s in range(sc_num):
            this_x = all_steps_data.iloc[sc_idxs[s], x_col_idx]
            this_y = all_steps_data.iloc[sc_idxs[s], y_col_idx]
            this_label = generate_sc_latency_label(
                all_steps_data, sc_idxs[s], sampling_rate, time_col_idx
            )
            ax[a].plot(this_x, this_y, label=this_label)
        # legend adjustments
        if legend_outside is True:
            ax[a].legend(
                fontsize=SC_LAT_LEGEND_FONTSIZE,
                loc="center left",
                bbox_to_anchor=(1, 0.5),
            )
        elif legend_outside is False:
            ax[a].legend(fontsize=SC_LAT_LEGEND_FONTSIZE)
        figure_file_string = " - " + angle + "Angle by Time"
        save_figures(f[a], results_dir, name, figure_file_string)
        if dont_show_plots:
            plt.close(f[a])

        # add figure to plot panel figures list
        if dont_show_plots is False:  # -> show plot panel
            plot_panel_instance.figures.append(f[a])


def plot_hindlimb_stickdiagram(all_steps_data, sc_idxs, info, cfg, plot_panel_instance):
    """4 - Plot a stick diagram of the hindlimb"""

    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    dont_show_plots = cfg["dont_show_plots"]
    convert_to_mm = cfg["convert_to_mm"]
    plot_joints = cfg["plot_joints"]
    sampling_rate = cfg["sampling_rate"]
    legend_outside = cfg["legend_outside"]
    color_palette = cfg["color_palette"]

    # some prep
    sc_num = len(sc_idxs)
    f, ax = plt.subplots(1, 1)
    color_cycle = plt.cycler("color", sns.color_palette(color_palette, sc_num))
    ax.set_prop_cycle(color_cycle)
    time_col_idx = all_steps_data.columns.get_loc(TIME_COL)

    # plot
    # => for timepoints from SC1 to SCend - plot(joint1x, joint1y)
    for s, this_color_dict in zip(range(sc_num), color_cycle):  # SC loop (colors)
        this_color = this_color_dict["color"][:3]
        this_label = generate_sc_latency_label(
            all_steps_data, sc_idxs[s], sampling_rate, time_col_idx
        )
        for i in sc_idxs[s]:  # loop over timepoints of current SC
            this_xs = list()  # for each timepoint, define joints' xy coord new
            this_ys = list()
            for joint in plot_joints:
                x_col_idx = all_steps_data.columns.get_loc(joint + "x")
                y_col_idx = all_steps_data.columns.get_loc(joint + "y")
                this_xs.append(all_steps_data.iloc[i, x_col_idx])
                this_ys.append(all_steps_data.iloc[i, y_col_idx])
            if i == sc_idxs[s][0]:
                ax.plot(this_xs, this_ys, color=this_color, label=this_label)
            else:
                ax.plot(this_xs, this_ys, color=this_color)
    ax.set_title(name + " - Primary Stick Diagram")
    # legend adjustments
    if legend_outside is True:
        ax.legend(
            fontsize=SC_LAT_LEGEND_FONTSIZE, loc="center left", bbox_to_anchor=(1, 0.5)
        )
    elif legend_outside is False:
        ax.legend(fontsize=SC_LAT_LEGEND_FONTSIZE)
    # labels & conversion
    ax.set_xlabel("x")  # will be used by conversion func if we convert
    ax.set_ylabel("y")
    if convert_to_mm:
        tickconvert_mm_to_cm(ax, "both")
    else:
        ax.set_xlabel("x (pixels)")
        ax.set_ylabel("y (pixels)")
    figure_file_string = " - Primary Stick Diagram"
    save_figures(f, results_dir, name, figure_file_string)
    if dont_show_plots:
        plt.close(f)

    # add figure to plot panel figures list
    if dont_show_plots is False:  # -> show plot panel
        plot_panel_instance.figures.append(f)


def plot_forelimb_stickdiagram(all_steps_data, sc_idxs, info, cfg, plot_panel_instance):
    """5 - Plot a stick diagram of the forelimb (for hindlimb stepcycles)"""

    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    dont_show_plots = cfg["dont_show_plots"]
    convert_to_mm = cfg["convert_to_mm"]
    fore_joints = cfg["fore_joints"]
    sampling_rate = cfg["sampling_rate"]
    legend_outside = cfg["legend_outside"]
    color_palette = cfg["color_palette"]

    # some prep
    sc_num = len(sc_idxs)
    f, ax = plt.subplots(1, 1)
    color_cycle = plt.cycler("color", sns.color_palette(color_palette, sc_num))
    ax.set_prop_cycle(color_cycle)
    time_col_idx = all_steps_data.columns.get_loc(TIME_COL)

    # plot
    for s, this_color in zip(range(sc_num), color_cycle):  # SC loop (colors)
        this_color = this_color["color"][:3]
        this_label = generate_sc_latency_label(
            all_steps_data, sc_idxs[s], sampling_rate, time_col_idx
        )
        for i in sc_idxs[s]:
            this_xs = list()
            this_ys = list()
            for joint in fore_joints:
                x_col_idx = all_steps_data.columns.get_loc(joint + "x")
                y_col_idx = all_steps_data.columns.get_loc(joint + "y")
                this_xs.append(all_steps_data.iloc[i, x_col_idx])
                this_ys.append(all_steps_data.iloc[i, y_col_idx])
            if i == sc_idxs[s][0]:
                ax.plot(this_xs, this_ys, color=this_color, label=this_label)
            else:
                ax.plot(this_xs, this_ys, color=this_color)
    ax.set_title(name + " - Secondary Stick Diagram")
    if convert_to_mm:
        tickconvert_mm_to_cm(ax, "both")
    # legend adjustments
    if legend_outside is True:
        ax.legend(
            fontsize=SC_LAT_LEGEND_FONTSIZE, loc="center left", bbox_to_anchor=(1, 0.5)
        )
    elif legend_outside is False:
        ax.legend(fontsize=SC_LAT_LEGEND_FONTSIZE)
    # labels & conversion
    ax.set_xlabel("x")  # will be used by conversion func if we convert
    ax.set_ylabel("y")
    if convert_to_mm:
        tickconvert_mm_to_cm(ax, "both")
    else:
        ax.set_xlabel("x (pixels)")
        ax.set_ylabel("y (pixels)")
    figure_file_string = " - Secondary Stick Diagram"
    save_figures(f, results_dir, name, figure_file_string)
    if dont_show_plots:
        plt.close(f)

    # add figure to plot panel figures list
    if dont_show_plots is False:  # -> show plot panel
        plot_panel_instance.figures.append(f)


def plot_joint_y_by_average_SC(average_data, std_data, info, cfg, plot_panel_instance):
    """6 - Plot joints' y as a function of average SC's percentage"""

    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    dont_show_plots = cfg["dont_show_plots"]
    convert_to_mm = cfg["convert_to_mm"]
    bin_num = cfg["bin_num"]
    plot_SE = cfg["plot_SE"]
    sc_num = cfg["sc_num"]
    hind_joints = cfg["hind_joints"]
    legend_outside = cfg["legend_outside"]
    color_palette = cfg["color_palette"]

    # plot
    f, ax = plt.subplots(1, 1)
    ax.set_prop_cycle(
        plt.cycler("color", sns.color_palette(color_palette, len(hind_joints)))
    )
    x = np.linspace(0, 100, bin_num)
    for joint in hind_joints:  # joint loop (lines)
        y_col_idx = average_data.columns.get_loc(joint + "y")
        this_y = average_data.iloc[:, y_col_idx]  # average & std_data share colnames
        if plot_SE:
            this_std = std_data.iloc[:, y_col_idx] / np.sqrt(sc_num)
        else:
            this_std = std_data.iloc[:, y_col_idx]
        ax.plot(x, this_y, label=joint)
        ax.fill_between(x, this_y - this_std, this_y + this_std, alpha=0.2)
    # legend adjustments
    if legend_outside is True:
        ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    elif legend_outside is False:
        ax.legend()
    ax.set_title(name + " - Joint Y over average step cycle")
    # labels & conversion
    ax.set_xlabel("Percentage")
    ax.set_ylabel("y")  # used by conversion func
    if convert_to_mm:
        tickconvert_mm_to_cm(ax, "y")
    else:
        ax.set_ylabel("y (pixels)")
    figure_file_string = " - Joint Y-coord.s over average step cycle"
    save_figures(f, results_dir, name, figure_file_string)
    if dont_show_plots:
        plt.close(f)

    # add figure to plot panel figures list
    if dont_show_plots is False:  # -> show plot panel
        plot_panel_instance.figures.append(f)


def plot_joint_x_by_average_SC(average_data, std_data, info, cfg, plot_panel_instance):
    """7 - Plot joints' x as a function of average SC's percentage"""

    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    dont_show_plots = cfg["dont_show_plots"]
    convert_to_mm = cfg["convert_to_mm"]
    bin_num = cfg["bin_num"]
    plot_SE = cfg["plot_SE"]
    sc_num = cfg["sc_num"]
    hind_joints = cfg["hind_joints"]
    legend_outside = cfg["legend_outside"]
    color_palette = cfg["color_palette"]

    # plot
    f, ax = plt.subplots(1, 1)
    ax.set_prop_cycle(
        plt.cycler("color", sns.color_palette(color_palette, len(hind_joints)))
    )
    x = np.linspace(0, 100, bin_num)
    for joint in hind_joints:  # joint loop (lines)
        x_col_idx = average_data.columns.get_loc(joint + "x")
        this_y = average_data.iloc[:, x_col_idx]  # average & std_data share colnames
        if plot_SE:
            this_std = std_data.iloc[:, x_col_idx] / np.sqrt(sc_num)
        else:
            this_std = std_data.iloc[:, x_col_idx]
        ax.plot(x, this_y, label=joint)
        ax.fill_between(x, this_y - this_std, this_y + this_std, alpha=0.2)
    # legend adjustments
    if legend_outside is True:
        ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    elif legend_outside is False:
        ax.legend()
    ax.set_title(name + " - Joint X over average step cycle")
    # labels & conversion
    ax.set_xlabel("Percentage")
    ax.set_ylabel("x")  # used by conversion func
    if convert_to_mm:
        tickconvert_mm_to_cm(ax, "y")
    else:
        ax.set_ylabel("x (pixels)")
    figure_file_string = " - Joint X-coord.s over average step cycle"
    save_figures(f, results_dir, name, figure_file_string)
    if dont_show_plots:
        plt.close(f)

    # add figure to plot panel figures list
    if dont_show_plots is False:  # -> show plot panel
        plot_panel_instance.figures.append(f)


def plot_angles_by_average_SC(average_data, std_data, info, cfg, plot_panel_instance):
    """8 - Plot Angles as a function of average SC's percentage"""

    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    dont_show_plots = cfg["dont_show_plots"]
    bin_num = cfg["bin_num"]
    plot_SE = cfg["plot_SE"]
    sc_num = cfg["sc_num"]
    angles = cfg["angles"]
    legend_outside = cfg["legend_outside"]
    color_palette = cfg["color_palette"]

    # plot
    f, ax = plt.subplots(1, 1)
    ax.set_prop_cycle(
        plt.cycler("color", sns.color_palette(color_palette, len(angles["name"])))
    )
    x = np.linspace(0, 100, bin_num)
    ax.set_title(name + " - Joint angles over average step cycle")
    ax.set_xlabel("Percentage")
    ax.set_ylabel("Angle (degree)")
    for angle in angles["name"]:  # angle loop (lines)
        y_col_idx = average_data.columns.get_loc(angle + "Angle")
        this_y = average_data.iloc[:, y_col_idx]  # average & std_data share colnames
        if plot_SE:
            this_std = std_data.iloc[:, y_col_idx] / np.sqrt(sc_num)
        else:
            this_std = std_data.iloc[:, y_col_idx]
        ax.plot(x, this_y, label=angle)
        ax.fill_between(x, this_y - this_std, this_y + this_std, alpha=0.2)
    # legend adjustments
    if legend_outside is True:
        ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    elif legend_outside is False:
        ax.legend()
    figure_file_string = " - Joint angles over average step cycle"
    save_figures(f, results_dir, name, figure_file_string)
    if dont_show_plots:
        plt.close(f)

    # add figure to plot panel figures list
    if dont_show_plots is False:  # -> show plot panel
        plot_panel_instance.figures.append(f)


def plot_x_velocities_by_average_SC(
    average_data, std_data, info, cfg, plot_panel_instance
):
    """9 - Plot x velocities as a function of average SC's percentage"""

    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    dont_show_plots = cfg["dont_show_plots"]
    convert_to_mm = cfg["convert_to_mm"]
    bin_num = cfg["bin_num"]
    sampling_rate = cfg["sampling_rate"]
    plot_SE = cfg["plot_SE"]
    sc_num = cfg["sc_num"]
    hind_joints = cfg["hind_joints"]
    legend_outside = cfg["legend_outside"]
    color_palette = cfg["color_palette"]

    # plot
    f, ax = plt.subplots(1, 1)
    ax.set_prop_cycle(
        plt.cycler("color", sns.color_palette(color_palette, len(hind_joints)))
    )
    x = np.linspace(0, 100, bin_num)
    ax.set_title(name + " - Joint velocities over average step cycle")
    for joint in hind_joints:  # joint loop (lines)
        y_col_idx = average_data.columns.get_loc(joint + "Velocity")
        this_y = average_data.iloc[:, y_col_idx]  # average & std_data share colnames
        if plot_SE:
            this_std = std_data.iloc[:, y_col_idx] / np.sqrt(sc_num)
        else:
            this_std = std_data.iloc[:, y_col_idx]
        ax.plot(x, this_y, label=joint)
        ax.fill_between(x, this_y - this_std, this_y + this_std, alpha=0.2)
    # legend adjustments
    if legend_outside is True:
        ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    elif legend_outside is False:
        ax.legend()
    ax.set_xlabel("Percentage")
    ax.set_ylabel(
        "Velocity (x in pixel / " + str(int((1 / sampling_rate) * 1000)) + "ms)"
    )
    if convert_to_mm:
        tickconvert_mm_to_cm(ax, "y")
        ax.set_ylabel(
            "Velocity (x in cm / " + str(int((1 / sampling_rate) * 1000)) + "ms)"
        )
    figure_file_string = " - Joint velocities over average step cycle"
    save_figures(f, results_dir, name, figure_file_string)
    if dont_show_plots:
        plt.close(f)

    # add figure to plot panel figures list
    if dont_show_plots is False:  # -> show plot panel
        plot_panel_instance.figures.append(f)


def plot_angular_velocities_by_average_SC(
    average_data, std_data, info, cfg, plot_panel_instance
):
    """10 - Plot angular velocities as a function of average SC's percentage"""

    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    dont_show_plots = cfg["dont_show_plots"]
    bin_num = cfg["bin_num"]
    sampling_rate = cfg["sampling_rate"]
    plot_SE = cfg["plot_SE"]
    sc_num = cfg["sc_num"]
    angles = cfg["angles"]
    legend_outside = cfg["legend_outside"]
    color_palette = cfg["color_palette"]

    # plot
    f, ax = plt.subplots(1, 1)
    ax.set_prop_cycle(
        plt.cycler("color", sns.color_palette(color_palette, len(angles["name"])))
    )
    x = np.linspace(0, 100, bin_num)
    ax.set_title(name + " - Angular velocities over average step cycle")
    ax.set_xlabel("Percentage")
    ax.set_ylabel("Velocity (degree / " + str(int((1 / sampling_rate) * 1000)) + "ms)")
    for angle in angles["name"]:  # angle loop (lines)
        y_col_idx = average_data.columns.get_loc(angle + "Angle Velocity")  # space
        this_y = average_data.iloc[:, y_col_idx]  # average & std_data share colnames
        if plot_SE:
            this_std = std_data.iloc[:, y_col_idx] / np.sqrt(sc_num)
        else:
            this_std = std_data.iloc[:, y_col_idx]
        ax.plot(x, this_y, label=angle)
        ax.fill_between(x, this_y - this_std, this_y + this_std, alpha=0.2)
    # legend adjustments
    if legend_outside is True:
        ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    elif legend_outside is False:
        ax.legend()
    figure_file_string = " - Angular velocities over average step cycle"
    save_figures(f, results_dir, name, figure_file_string)
    if dont_show_plots:
        plt.close(f)

    # add figure to plot panel figures list
    if dont_show_plots is False:  # -> show plot panel
        plot_panel_instance.figures.append(f)


def plot_x_acceleration_by_average_SC(
    average_data, std_data, info, cfg, plot_panel_instance
):
    """11 - (optional) Plot x acceleration as a function of average SC's percentage"""

    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    dont_show_plots = cfg["dont_show_plots"]
    convert_to_mm = cfg["convert_to_mm"]
    bin_num = cfg["bin_num"]
    sampling_rate = cfg["sampling_rate"]
    plot_SE = cfg["plot_SE"]
    sc_num = cfg["sc_num"]
    hind_joints = cfg["hind_joints"]
    legend_outside = cfg["legend_outside"]
    color_palette = cfg["color_palette"]

    # plot
    f, ax = plt.subplots(1, 1)
    ax.set_prop_cycle(
        plt.cycler("color", sns.color_palette(color_palette, len(hind_joints)))
    )
    x = np.linspace(0, 100, bin_num)
    ax.set_title(name + " - Joint accelerations over average step cycle")
    for joint in hind_joints:  # joint loop (lines)
        y_col_idx = average_data.columns.get_loc(joint + "Acceleration")  # no space
        this_y = average_data.iloc[:, y_col_idx]  # average & std_data share colnames
        if plot_SE:
            this_std = std_data.iloc[:, y_col_idx] / np.sqrt(sc_num)
        else:
            this_std = std_data.iloc[:, y_col_idx]
        ax.plot(x, this_y, label=joint)
        ax.fill_between(x, this_y - this_std, this_y + this_std, alpha=0.2)
    # legend adjustments
    if legend_outside is True:
        ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    elif legend_outside is False:
        ax.legend()
    ax.set_xlabel("Percentage")
    ax.set_ylabel(
        "Acceleration (x in pixel / " + str(int((1 / sampling_rate) * 1000)) + "ms)"
    )
    if convert_to_mm:
        tickconvert_mm_to_cm(ax, "y")
        ax.set_ylabel(
            "Acceleration (x in cm / " + str(int((1 / sampling_rate) * 1000)) + "ms)"
        )
    figure_file_string = " - Joint acceleration over average step cycle"
    save_figures(f, results_dir, name, figure_file_string)
    if dont_show_plots:
        plt.close(f)

    # add figure to plot panel figures list
    if dont_show_plots is False:  # -> show plot panel
        plot_panel_instance.figures.append(f)


def plot_angular_acceleration_by_average_SC(
    average_data, std_data, info, cfg, plot_panel_instance
):
    """12 - (optional) Plot angular acceleration as a function of average SC's
    percentage
    """

    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    dont_show_plots = cfg["dont_show_plots"]
    bin_num = cfg["bin_num"]
    sampling_rate = cfg["sampling_rate"]
    plot_SE = cfg["plot_SE"]
    sc_num = cfg["sc_num"]
    angles = cfg["angles"]
    legend_outside = cfg["legend_outside"]
    color_palette = cfg["color_palette"]

    # plot
    f, ax = plt.subplots(1, 1)
    ax.set_prop_cycle(
        plt.cycler("color", sns.color_palette(color_palette, len(angles["name"])))
    )
    x = np.linspace(0, 100, bin_num)
    ax.set_title(name + " - Angular accelerations over average step cycle")
    ax.set_xlabel("Percentage")
    ax.set_ylabel(
        "Acceleration (degree / " + str(int((1 / sampling_rate) * 1000)) + "ms)"
    )
    for angle in angles["name"]:  # angle loop (lines)
        y_col_idx = average_data.columns.get_loc(angle + "Angle Acceleration")
        this_y = average_data.iloc[:, y_col_idx]  # average & std_data share colnames
        if plot_SE:
            this_std = std_data.iloc[:, y_col_idx] / np.sqrt(sc_num)
        else:
            this_std = std_data.iloc[:, y_col_idx]
        ax.plot(x, this_y, label=angle)
        ax.fill_between(x, this_y - this_std, this_y + this_std, alpha=0.2)
    # legend adjustments
    if legend_outside is True:
        ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    elif legend_outside is False:
        ax.legend()
    figure_file_string = " - Angular acceleration over average step cycle"
    save_figures(f, results_dir, name, figure_file_string)
    if dont_show_plots:
        plt.close(f)

    # add figure to plot panel figures list
    if dont_show_plots is False:  # -> show plot panel
        plot_panel_instance.figures.append(f)


def save_figures(figure, results_dir, name, figure_file_string):
    """Save figures as pngs to results_dir and as svgs to separate subfolders"""
    # pngs to results_dir
    figure.savefig(
        os.path.join(results_dir, name + figure_file_string + ".png"),
        bbox_inches="tight",
    )
    # svgs to subfolders
    svg_dir = os.path.join(results_dir, "SVG Figures")
    if not os.path.exists(svg_dir):
        os.makedirs(svg_dir)
    figure.savefig(
        os.path.join(svg_dir, name + figure_file_string + ".svg"), bbox_inches="tight"
    )


def tickconvert_mm_to_cm(axis, whichlabel):
    """Convert axis-ticks from mm (of data) to cm"""
    if (whichlabel == "both") | (whichlabel == "x"):
        x_ticks = axis.get_xticks()
        x_ticklabels = []
        for t in x_ticks:
            x_ticklabels.append(str(round(t / 10, 2)))
        axis.set_xticks(x_ticks, labels=x_ticklabels)
        old_xlabel = axis.get_xlabel()
        axis.set_xlabel(old_xlabel + " (cm)")
    if (whichlabel == "both") | (whichlabel == "y"):
        y_ticks = axis.get_yticks()
        y_ticklabels = []
        for t in y_ticks:
            y_ticklabels.append(str(round(t / 10, 2)))
        axis.set_yticks(y_ticks, labels=y_ticklabels)
        old_ylabel = axis.get_ylabel()
        axis.set_ylabel(old_ylabel + " (cm)")


def generate_sc_latency_label(all_steps_data, this_sc_idx, sampling_rate, time_col_idx):
    if sampling_rate <= 100:
        float_precision = 2  # how many decimals we round to
    else:
        float_precision = 4
    this_label = (
        str(
            round(
                all_steps_data.iloc[this_sc_idx[0], time_col_idx],
                float_precision,
            )
        )
        + "-"
        + str(
            round(
                all_steps_data.iloc[this_sc_idx[-1], time_col_idx],
                float_precision,
            )
        )
        + "s"
    )
    return this_label



# %% imports
from autogaita.common2D.common2D_constants import FILE_ID_STRING_ADDITIONS
from autogaita.dlc.dlc_utils import find_number
from autogaita.resources.utils import try_to_run_gaita, write_issues_to_textfile
import os
import numpy as np
import tkinter as tk


def run_singlerun_in_multirun(tracking_software, idx, info, folderinfo, cfg):
    """When performing a multirun, either via Batch Analysis in GUI or batchrun scripts, run the analysis for a given dataset"""
    # extract and pass info of this mouse/run or name (also update resdir)
    this_info = {}
    keynames = info.keys()
    for keyname in keynames:
        this_info[keyname] = info[keyname][idx]
        if cfg["results_dir"]:
            this_info["results_dir"] = os.path.join(
                cfg["results_dir"], this_info["name"]
            )
        else:
            this_info["results_dir"] = os.path.join(
                folderinfo["root_dir"], "Results", this_info["name"]
            )
    # important to only pass this_info to main script here (1 run at a time!)
    try_to_run_gaita(tracking_software, this_info, folderinfo, cfg, True)


def extract_info(tracking_software, folderinfo, in_GUI=False):
    """Prepare a dict of lists that include unique infos for each dataset in a folder"""

    # DLC
    if tracking_software == "DLC":
        # unpack
        premouse_string = folderinfo["premouse_string"]
        postmouse_string = folderinfo["postmouse_string"]
        prerun_string = folderinfo["prerun_string"]
        postrun_string = folderinfo["postrun_string"]

        # prepare output info dict and run
        info = {"name": [], "mouse_num": [], "run_num": []}
        for filename in os.listdir(folderinfo["root_dir"]):
            # make sure we don't get wrong files
            if (
                (premouse_string in filename)
                & (prerun_string in filename)
                & (filename.endswith(".csv"))
            ):
                # ID number - fill in "_" for user if needed
                this_mouse_num = False
                for string_addition in FILE_ID_STRING_ADDITIONS:
                    try:
                        candidate_postmouse_string = string_addition + postmouse_string
                        this_mouse_num = find_number(
                            filename,
                            premouse_string,
                            candidate_postmouse_string,
                        )
                    except:
                        pass
                if this_mouse_num is False:
                    no_ID_num_found_msg = (
                        "Unable to extract ID numbers from file identifiers! "
                        + "Check unique subject [B] and unique task [C] identifiers"
                    )
                    # if we're in the GUI, show an error message & stop here
                    if in_GUI:
                        tk.messagebox.showerror(
                            title="No ID number found!", message=no_ID_num_found_msg
                        )
                        return
                # Do the same for run number
                this_run_num = False
                for string_addition in FILE_ID_STRING_ADDITIONS:
                    try:
                        candidate_postrun_string = string_addition + postrun_string
                        this_run_num = find_number(
                            filename, prerun_string, candidate_postrun_string
                        )
                    except:
                        pass
                if this_run_num is False:
                    no_run_num_found_msg = (
                        "Unable to extract trial numbers from file identifiers! "
                        + "Check unique trial [D] and unique camera [E] identifiers"
                    )
                    if in_GUI:
                        tk.messagebox.showerror(
                            title="No Trial number found!", message=no_run_num_found_msg
                        )
                        return
                # if we found both an ID and a run number, create this_name & add to dict
                if (
                    this_mouse_num and this_run_num
                ):  # are truthy since not False if found
                    this_name = (
                        "ID " + str(this_mouse_num) + " - Run " + str(this_run_num)
                    )
                    if this_name not in info["name"]:  # no data/beam duplicates here
                        info["name"].append(this_name)
                        info["mouse_num"].append(this_mouse_num)
                        info["run_num"].append(this_run_num)

    # SLEAP
    elif tracking_software == "SLEAP":
        # unpack
        root_dir = folderinfo["root_dir"]
        data_string = folderinfo["data_string"]
        # prepare output info dict and run
        info = {"name": []}
        for filename in os.listdir(root_dir):
            if data_string + ".csv" in filename:
                info["name"].append(filename.split(data_string + ".csv")[0])
            if data_string + ".h5" in filename:
                info["name"].append(filename.split(data_string + ".h5")[0])
        info["name"] = list(set(info["name"]))  # no duplicates

    return info


# ...........................  SC extraction helpers  ..................................
def check_cycle_out_of_bounds(all_cycles):
    """Check if user provided SC latencies that were not in video/data bounds"""
    clean_cycles = None
    for cycle in all_cycles:
        # below checks if values are any type of int (just in case this should
        # for some super random reason change...)
        if isinstance(cycle[0], (int, np.integer)) & isinstance(
            cycle[1], (int, np.integer)
        ):
            if clean_cycles is None:
                clean_cycles = []
            clean_cycles.append(cycle)
    return clean_cycles


def check_cycle_duplicates(all_cycles):
    """Check if there are any duplicate SC latencies.
    This would break our plotting functions, which use .loc on all_steps_data - thus,
    all indices of all_cycles have to be unique. If any duplicates found, add one
    datapoint to the start latency.
    """
    for c, cycle in enumerate(all_cycles):
        if c > 0:
            if cycle[0] == all_cycles[c - 1][1]:
                all_cycles[c][0] += 1
    return all_cycles


def check_cycle_order(all_cycles, info):
    """Check if user input flawed SC latencies

    Two cases
    1. Start latency earlier than end latency of previous SC
    2. End latency earlier then start latency of current SC
    """
    clean_cycles = []
    current_max_time = 0
    for c, cycle in enumerate(all_cycles):
        if cycle[0] > current_max_time:
            if cycle[1] > cycle[0]:
                clean_cycles.append(cycle)  # only append if both tests passed
                current_max_time = cycle[1]
            else:
                this_message = (
                    "\n***********\n! WARNING !\n***********\n"
                    + "SC #"
                    + str(c + 1)
                    + " has a later start than end latency - Skipping!"
                )
                print(this_message)
                write_issues_to_textfile(this_message, info)
        else:
            this_message = (
                "\n***********\n! WARNING !\n***********\n"
                + "SC #"
                + str(c + 1)
                + " has an earlier start than previous SC's end latency - Skipping!"
            )
            print(this_message)
            write_issues_to_textfile(this_message, info)
    return clean_cycles


def check_tracking(data, info, all_cycles, cfg):
    """Check if any x/y column of any joint has broken datapoints"""
    # unpack
    convert_to_mm = cfg["convert_to_mm"]
    x_sc_broken_threshold = cfg["x_sc_broken_threshold"]
    y_sc_broken_threshold = cfg["y_sc_broken_threshold"]
    pixel_to_mm_ratio = cfg["pixel_to_mm_ratio"]
    hind_joints = cfg["hind_joints"]
    if convert_to_mm:
        x_sc_broken_threshold = x_sc_broken_threshold / pixel_to_mm_ratio
        y_sc_broken_threshold = y_sc_broken_threshold / pixel_to_mm_ratio
    columns = []
    clean_cycles = None
    for joint in hind_joints:
        columns.append(joint + "x")
        columns.append(joint + "y")
    for c, cycle in enumerate(all_cycles):
        exclude_this_cycle = False  # reset
        for col in columns:
            if col.endswith("x"):
                this_threshold = x_sc_broken_threshold
            elif col.endswith("y"):
                this_threshold = y_sc_broken_threshold
            this_data = data.loc[cycle[0] : cycle[1], col]
            for i in range(len(this_data) - 1):
                if (this_data.iloc[i + 1] > (this_data.iloc[i] + this_threshold)) | (
                    this_data.iloc[i + 1] < (this_data.iloc[i] - this_threshold)
                ):
                    exclude_this_cycle = True
        if exclude_this_cycle == True:
            this_message = (
                "\n...excluding SC #" + str(c + 1) + " - DLC tracking failed!"
            )
            print(this_message)
            write_issues_to_textfile(this_message, info)
        else:
            if clean_cycles == None:
                clean_cycles = [cycle]  # also makes a 2xscs list of lists
            else:
                clean_cycles.append(cycle)
    return clean_cycles


def handle_issues(condition, info):
    """Handle different kind of issues with step-cycles (& the table)"""
    # 1: can also occur bc. all scs when dlc failed
    if condition == "scs_invalid":
        this_message = (
            "\n***********\n! WARNING !\n***********\n"
            + "Skipped since all SCs invalid!"
        )
        print(this_message)
        write_issues_to_textfile(this_message, info)
    # 2: no SCs were provided in XLS table
    elif condition == "no_scs":
        this_message = (
            "\n***********\n! WARNING !\n***********\n"
            + "Skipped since no SCs in Annotation Table!"
        )
        print(this_message)
        write_issues_to_textfile(this_message, info)
    # 3: the mouse was not included in XLS table
    elif condition == "no_mouse":
        this_message = (
            "\n***********\n! WARNING !\n***********\n"
            + "Skipped since ID not in Annotation Table!"
        )
        print(this_message)
        write_issues_to_textfile(this_message, info)
    # 4: user entered wrong column names in XLS table
    elif condition == "wrong_scxls_colnames":
        this_message = (
            "\n******************\n! CRITICAL ERROR !"
            + "\n******************\n"
            + "Annotation Table's Column Names are wrong!\n"
            + "Check Instructions!"
        )
        print(this_message)
        write_issues_to_textfile(this_message, info)
    # 5:
    elif condition == "double_mouse":
        this_message = (
            "\n***********\n! WARNING !\n***********\n"
            + "Skipped since ID found more than once in "
            + "Annotation Table!"
        )
        print(this_message)
        write_issues_to_textfile(this_message, info)
    return
